/*Interface between the C and Pascal parts of GP

  Copyright (C) 2000-2006 Free Software Foundation, Inc.

  Author: Frank Heckenbach <frank@pascal.gnu.de>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, version 2.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING. If not, write to
  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA. */

/* gp.pas */

typedef enum { fk_None, fk_Program, fk_Unit, fk_Module, fk_CFile, fk_Object } TFileKind;

typedef struct CStringList
{
  struct CStringList *Next;
  char *Name1, *Name2;
} CStringList;

void Error (const char *Message) __attribute__ ((noreturn));
void ErrorFile (const char *FileName, const char *Message) __attribute__ ((noreturn));
void ErrorLineColumnCString (const char *FileName, int Line, int Column, const char *Msg) __attribute__ ((noreturn));
void WarningFile (const char *FileName, const char *Message);
void WarningLineColumnCString (const char *FileName, int Line, int Column, const char *Msg);

/* gp-parse.l */

extern const char *gpc_options[];

TFileKind ParseFile (char *_SrcName, char *_Buf, int _BufCount, int *_ImplementationPos,
                     char **_ModuleName, CStringList **_ExportedInterfaces,
                     CStringList **_InterfaceUses, CStringList **_ImplementationUses,
                     CStringList **_LinkedFiles, CStringList **_ResourceFiles);

#ifdef V08

/* cfg-lex.c */

int yylex (void);

/* cfg-parse.y */

void yyerror (const char *) __attribute__ ((noreturn));
void yywarning (const char *);
int yyparse (void);
void lex_error (const char *, int);

/* cfg.pas */

typedef struct String *PString, *PID;
typedef struct TIDList TIDList;
typedef struct TLocalDecl TLocalDecl;
typedef struct TLocalDeclList TLocalDeclList;
typedef struct TStmtList TStmtList;
typedef struct TCond TCond;
typedef enum { OP_NONE, OP_NOT, OP_AND, OP_OR } TBoolOp;
typedef enum { OP_SEQ, OP_SLT, OP_SGT, OP_SNE, OP_SGE, OP_SLE, OP_EQ, OP_GT, OP_LT, OP_NE, OP_GE, OP_LE } TCompOp;
typedef struct TExpr TExpr;
typedef struct TExprList TExprList;
typedef struct TStrConstList TStrConstList;

PString CStringToPString (const char *);
void ExecuteStmtsOnce (TStmtList *);
void AddFinalizer (TStmtList *);
void StartRoutine (PID, PID, TLocalDeclList *, TLocalDeclList *, int);
void FinishRoutine (TStmtList *);
TIDList *AppendIDList (TIDList *, PID);
TLocalDecl *MakeLocalDecl (TIDList *, TExpr *, int);
TLocalDeclList *AppendLocalDeclList (TLocalDeclList *, TLocalDecl *);
TLocalDeclList *ConcatLocalDeclList (TLocalDeclList *, TLocalDeclList *);
TStmtList *ConcatStmtList (TStmtList *, TStmtList *);
void DeclareVars (TLocalDeclList *);
void CheckIDAssign (PID);
TStmtList *MakeAssignStmt (PID, TExpr *);
TStmtList *MakeIfStmt (TCond *, TStmtList *, TStmtList *);
TStmtList *MakeWhileStmt (TCond *, TStmtList *);
TStmtList *MakeRepeatStmt (TCond *, TStmtList *);
TStmtList *MakeCallStmt (PID, TExprList *);
TCond *MakeCondExpr (TBoolOp, TCond *, TCond *);
TCond *MakeComparison (TCompOp, TExpr *, TExpr *);
TExpr *MakeIDExpr (PID);
TExpr *MakeCallExpr (PID, TExprList *);
TExpr *MakeStrConstListExpr (TStrConstList *);
TExpr *MakeListExpr (TExprList *);
TExpr *MakeConcatExpr (TExpr *, TExpr *);
TExprList *AppendExprList (TExprList *, TExpr *);
TStrConstList *AppendStrConstList (TStrConstList *, PString);
#endif
