{ GP, the GPC make utility

  Copyright (C) 2000-2006 Free Software Foundation, Inc.

  Author: Frank Heckenbach <frank@pascal.gnu.de>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, version 2.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING. If not, write to
  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA. }

{$gnu-pascal,I+}
{$if __GPC_RELEASE__ < 20041017}
{$error GP requires GPC release 20041017 or newer.}
{$endif}

program GP;

uses GPC, MD5, Pipes, StringUtils, FileUtils;

{$ifdef V08}

uses Cfg;

procedure GPConfigInit (var PC, CC, ExecutableSuffix: String; var Options: PStringListVal);

  function GetPlatform: TString;
  begin
    GetPlatform := ''  { @@ }
  end;

begin
  InitializeParseConfigFiles (GetPlatform);
  ParseConfigFile ( @@ 'init.gp');
  FinalizeParseConfigFiles (PC, CC, ExecutableSuffix, Options)
end;

{$else}

type
  PStringListVal = ^TStringListVal;
  TStringListVal (Count: Integer) = record
    Elements: array [1 .. Max (Count, 1)] of PString
  end;

procedure GPConfigInit (var PC, CC, ExecutableSuffix: String; var Options: PStringListVal);
var i: Integer;
begin
  CC := GetEnv ('CC');
  PC := GetEnv ('PC');
  ExecutableSuffix := '';
  New (Options, ParamCount);
  for i := 1 to Options^.Count do
    Options^.Elements[i] := NewString (ParamStr (i))
end;

{$endif}

type
  TFileKind = (fk_None, fk_Program, fk_Unit, fk_Module, fk_CFile, fk_Object);
  PCStringList = ^TCStringList;
  TCStringList = record
    Next: PCStringList;
    Name1, Name2: CString
  end;

const
  NormalizedKind: array [TFileKind] of TFileKind =
    (fk_None, fk_Program, fk_Unit, fk_Unit, fk_CFile, fk_Object);

  LanguageNormalizedKind: array [TFileKind] of TFileKind =
    (fk_None, fk_Program, fk_Program, fk_Program, fk_CFile, fk_Object);

  LanguageSuffixes: array [TFileKind, 1 .. 17] of String (4) =
    (('', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', ''),
     ('.pas', '.p', '.pp', '.dpr', '', '', '', '', '', '', '', '', '', '', '', '', ''),
     ('.pas', '.p', '.pp', '.dpr', '', '', '', '', '', '', '', '', '', '', '', '', ''),
     ('.pas', '.p', '.pp', '.dpr', '', '', '', '', '', '', '', '', '', '', '', '', ''),
     ('.c', '.cc', '.cpp', '.CPP', '.cp', '.C', '.c++', '.cxx', '.i', '.ii', '.m', '.mi', '.mm', '.M', '.mii', '.S', '.s'),
     ('.o', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', ''));

var
  GPCOptions: array [0 .. MaxVarSize div SizeOf (CString) - 1] of CString; external name 'gpc_options';

{$L gp-parse.c}
function ParseFile (SrcName: CString; var Buf; BufCount: CInteger; var ImplementationPos: CInteger; var ModuleName: CString;
                    var ExportedInterfaces, InterfaceUses, ImplementationUses, LinkedFiles, ResourceFiles: PCStringList): TFileKind; external name 'ParseFile';

const
  ProgramVersion = {$ifdef RELEASE_VERSION} 'version ' + RELEASE_VERSION {$else} 'unknown version' {$endif};
  GPDHeader: array [Boolean] of String (49) =
    ('GNU Pascal Dependency File, Version 2',
     'GNU Pascal Preliminary Dependency File, Version 2');
  DefaultUnitDestinationPath = DirSelf + DirSeparator;

  { Preprocessed C source files are usually vastly blown up with hot air
    (a.k.a. header files), so let's count their lines in a more realistic way. ;-}
  CLinesDiv = 10;

type
  PBytes = ^TBytes;
  TBytes (Size: Integer) = array [0 .. Size - 1] of Byte;

  PPStringList = ^PStringList;
  PStringList = ^TStringList;
  TStringList = record
    Next: PStringList;
    s: TString
  end;

  TRecompileRelevant = (rr_None, rr_Link, rr_Compile);
  PPOptionList = ^POptionList;
  POptionList = ^TOptionList;
  TOptionList = record
    Next: POptionList;
    Option: TString;
    RecompileRelevant: TRecompileRelevant
  end;

  PPUnitList = ^PUnitList;
  PUnitList = ^TUnitList;

  PDep = ^TDep;

  PPDepList = ^PDepList;
  PDepList = ^TDepList;
  TDepList = record
    Next: PDepList;
    Dep: PDep
  end;

  TDep = record
    Kind: TFileKind;
    SrcName, SrcBaseName, SrcBaseNameExt, ModuleName, InterfaceName, ObjectBasePath, ObjectName,
      GPIBasePath, GPIName, GPDName, Platform, Version, Options: TString;
    ExportedInterfaces, Includes, LinkedLibs, LinkedObjects, LinkedFiles: PStringList;
    Lines, PBLines, Number, LowLink: Integer;
    StackNext: PDep;
    Main, ApplyAutoUses: Boolean;
    UsedUnits: array [Boolean] of PUnitList;  { index means `interface' }
    UsedBy: array [Boolean] of PDepList;      { index means `interface' }
    MD5: TMD5;
    MD5Units: array [Boolean] of TMD5;        { index means `interface' }
    Recompile: (rc_None, rc_Impossible, rc_SrcChanged, rc_AlwaysButGPIOK, rc_Always);
    InDegree: array [Boolean] of Integer;     { index means `interface' }
    CyclePosition: PDepList;
    DepsProcessed: (dp_None, dp_Interface, dp_All);
    InterfaceUsedUnitRecompiled, LinkOptionsMismatch: Boolean
  end;

  TUnitList = record
    Next: PUnitList;
    FileName, InterfaceName: TString;
    MD5: TMD5;
    Dep: PDep
  end;

  TOptionSelection = (co_OnlyVersion, co_NoVerbose, co_Compile, co_All);
  TGPD = (gd_None, gd_Preliminary, gd_Final);

var
  ProgramName, MainSrc, AutoUses, AutoUsesOption, UnitPath, ObjectPath,
    UnitDestinationPath, ObjectDestinationPath, ExecutableSuffix, MainDestName: TString;
  LinkFlag, AssembleFlag, TestParserFlag, AutoBuild, MkDirFlag, NoWarnings,
    DefaultPaths, MainDestNameGiven, CheckPlatform, CheckVersion, CheckOptions,
    ProgressMessageInSubprocess, ProgressBarInSubprocess, ProgressMessages, ProgressBar,
    PrintOnly, PrintPlatform, PrintSources, PrintObjects, PrintLibs, PrintExecutable,
    PrintCycles: Boolean;
  NeededOptions: PPStrings = nil;
  CmdLineOptions: PStringListVal;
  AutoUsedUnits: PUnitList = nil;
  Verbosity, Cleaning, TotalLines, LinesDone: Integer;
  LinkOptionsStr: TString;
  Compilers: array [Boolean] of record
    Initialized: Boolean;
    OptionCount: array [TOptionSelection] of Integer;
    Options: array [TOptionSelection] of POptionList;
    ProgName, Platform, Version, OptionStr: TString
  end;

{ Error and message handling }

procedure ProgressMessageStartSubprocess;
begin
  if ProgressMessages then
    begin
      if not ProgressMessageInSubprocess then WriteLn (StdErr, #1'#progress+#');
      ProgressMessageInSubprocess := True
    end
end;

procedure ProgressMessage (const Message: String);
begin
  if ProgressMessages then
    begin
      if ProgressMessageInSubprocess then WriteLn (StdErr, #1'#progress-#');
      ProgressMessageInSubprocess := False;
      WriteLn (StdErr, #1'#progress# ', Message)
    end
end;

procedure ProgressBarDel;
begin
  if ProgressBar and ProgressBarInSubprocess then
    begin
      WriteLn (StdErr, #1'#progress-bar-#');
      ProgressBarInSubprocess := False
    end
end;

procedure ProgressBarSet (const Position: Real);
begin
  if ProgressBar then
    begin
      ProgressBarDel;
      WriteLn (StdErr, #1'#progress-bar# ', Position)
    end
end;

procedure ProgressBarAdd (const Offset, Size, Range: Real);
begin
  if ProgressBar then
    begin
      ProgressBarDel;
      WriteLn (StdErr, #1'#progress-bar+#', Offset, ', ', Size, ', ', Range);
      ProgressBarInSubprocess := True
    end
end;

procedure Quit (Status: Integer); attribute (noreturn);
begin
  ProgressMessage ('');
  if Status = 0 then ProgressBarSet (1);
  Halt (Status)
end;

procedure PrintMessage (const Prefix, Message: String);
var i, j: Integer;
begin
  i := 0;
  repeat
    j := PosFrom (NewLine, Message, i + 1);
    if j = 0 then j := Length (Message) + 1;
    WriteLn (StdErr, Prefix, '' : Ord (i > 0), Copy (Message, i + 1, j - i - 1));
    i := j
  until i > Length (Message)
end;

procedure GeneralError (const Prefix, Message: String); attribute (noreturn);
begin
  PrintMessage (Prefix, Message);
  Quit (1)
end;

procedure Error (const Message: CString); attribute (noreturn, name = 'Error');
begin
  GeneralError (ProgramName + ': ', CString2String (Message))
end;

procedure ErrorFile (const FileName, Message: CString); attribute (noreturn, name = 'ErrorFile');
begin
  GeneralError (CString2String (FileName) + ': ', CString2String (Message))
end;

{$ifdef V08}
procedure ErrorLine (const FileName: String; Line: Integer; const Message: String); attribute (noreturn);
var Prefix: TString;
begin
  WriteStr (Prefix, FileName, ':', Line, ': ');
  GeneralError (Prefix, Message)
end;
{$endif}

procedure ErrorLineColumnCString (FileName: CString; Line, Column: Integer; Msg: CString); attribute (noreturn, name = 'ErrorLineColumnCString');
var Prefix: TString;
begin
  WriteStr (Prefix, CString2String (FileName), ':', Line, ':', Column, ': ');
  GeneralError (Prefix, CString2String (Msg))
end;

procedure Message (Level: Integer; const Message: String);
begin
  if Verbosity >= Level then WriteLn (StdErr, Message)
end;

procedure GeneralWarning (const Prefix, Message: String);
begin
  PrintMessage (Prefix + 'warning: ', Message)
end;

procedure Warning (const Message: String);
begin
  GeneralWarning (ProgramName + ': ', Message)
end;

procedure WarningFile (const FileName, Message: CString); attribute (name = 'WarningFile');
begin
  GeneralWarning (CString2String (FileName) + ': ', CString2String (Message))
end;

procedure WarningLineColumnCString (FileName: CString; Line, Column: Integer; Msg: CString); attribute (name = 'WarningLineColumnCString');
var Prefix: TString;
begin
  WriteStr (Prefix, CString2String (FileName), ':', Line, ':', Column, ': ');
  GeneralWarning (Prefix, CString2String (Msg))
end;

{ Utilities }

function Ext2Language (const Ext: String) = Res: TFileKind;
var i: Integer;
begin
  for Res := Low (Res) to High (Res) do
    for i := Low (LanguageSuffixes[Res]) to High (LanguageSuffixes[Res]) do
      if StrEqualCase (LanguageSuffixes[Res, i], Ext) then Exit;
  Res := fk_None
end;

{ Returns -1 if file does not exist }
function GetModificationTime (const aFileName: String): UnixTimeType;
var
  b: BindingType;
  f: Text;
begin
  GetModificationTime := -1;
  if aFileName = '' then Exit;
  Assign (f, aFileName);
  b := Binding (f);
  Unbind (f);
  if b.Bound and b.Existing then GetModificationTime := b.ModificationTime
end;

procedure CreateDir (const FileName: String); attribute (iocritical);
var
  b: BindingType;
  f: Text;
begin
  if MkDirFlag then MkDirs (FileName);
  Assign (f, FileName);
  b := Binding (f);
  Unbind (f);
  if not (b.Bound and b.Directory and b.Writable) then
    ErrorFile (FileName, 'no directory or not writable')
end;

{ GPD file handling }

function ReadGPD (var Dep: TDep): TGPD;
var
  f: Text;
  LineNumber, OrdKind, ValError: Integer;
  Line, s: TString;
  Preliminary, PascalFlag: Boolean;
  PExportedInterfaces, PIncludes, PLinkedLibs, PLinkedObjects, PLinkedFiles: PPStringList;
  PInterfaceUses, PImplementationUses: PPUnitList;

  procedure ReadGPDError (const Message: String);
  begin
    WriteLn (StdErr, Dep.GPDName, ':', LineNumber, ': ', Message);
    WriteLn (StdErr, 'ignoring ', Dep.GPDName);
    Close (f)
  end;

  procedure NextLine;
  begin
    Inc (LineNumber);
    ReadLn (f, Line)
  end;

  function ReadLine (const Prefix: String; var s: String): Boolean;
  begin
    ReadLine := False;
    NextLine;
    if not IsPrefix (Prefix, Line) then
      begin
        ReadGPDError ('`' + Prefix + ''' expected (ignoring)');
        Exit
      end;
    s := Copy (Line, Length (Prefix) + 1);
    ReadLine := True
  end;

  function ReadMD5 (const Prefix: String; var MD5: TMD5): Boolean;
  var s: TString;
  begin
    ReadMD5 := False;
    if not ReadLine (Prefix, s) then Exit;
    if not MD5Val (s, MD5) then
      begin
        ReadGPDError ('invalid MD5 value');
        Exit
      end;
    ReadMD5 := True
  end;

  function AddStringList (Allowed: Boolean; var p: PPStringList): Boolean;
  begin
    AddStringList := False;
    if not Allowed then
      begin
        ReadGPDError ('invalid linking directive for C file');
        Exit
      end;
    New (p^);
    p^^.s := Copy (Line, 3);
    p := @p^^.Next;
    AddStringList := True
  end;

  function AddUnitList (var p: PPUnitList): Boolean;
  var i, j: Integer;
  begin
    AddUnitList := False;
    if not PascalFlag then
      begin
        ReadGPDError ('invalid unit dependency for C file');
        Exit
      end;
    New (p^);
    i := PosFrom (' ', Line, 3);
    j := LastPos (' ', Line);
    if (i >= j) or not MD5Val (Copy (Line, j + 1), p^^.MD5) then
      begin
        ReadGPDError ('invalid unit entry');
        Exit
      end;
    p^^.InterfaceName := Copy (Line, 3, i - 3);
    p^^.FileName := Copy (Line, i + 1, j - i - 1);
    p^^.Dep := nil;
    p := @p^^.Next;
    AddUnitList := True
  end;

begin
  ReadGPD := gd_None;
  {$local I-}
  Reset (f, Dep.GPDName);
  {$endlocal}
  if IOResult <> 0 then Exit;
  LineNumber := 0;
  with Dep do
    begin
      if not ReadLine ('', s) then Exit;
      Preliminary := False;
      if s = GPDHeader[True] then
        Preliminary := True
      else if s <> GPDHeader[False] then
        begin
          ReadGPDError ('header not found');
          Exit
        end;
      if not ReadLine ('T ', s) then Exit;
      Val (s, OrdKind, ValError);
      if (ValError <> 0) or (OrdKind < Ord (Low (Kind))) or (OrdKind > Ord (High (Kind))) then
        begin
          ReadGPDError ('invalid entry');
          Exit
        end;
      Kind := TFileKind (OrdKind);
      PascalFlag := Kind <> fk_CFile;
      if not ReadLine ('t ', Platform) then Exit;
      if not ReadLine ('V ', Version) then Exit;
      if not ReadLine ('P ', Options) then Exit;
      if Kind = fk_Program then
        if ReadLine ('p ', s) then
          LinkOptionsMismatch := s <> LinkOptionsStr
        else
          Exit;
      if not Preliminary then
        begin
          if not ReadLine ('s ', SrcName) then Exit;
          if not ReadLine ('O ', ObjectName) then Exit
        end;
      if not ReadLine ('g ', GPIName) then Exit;
      if not ReadLine ('n ', s) then Exit;
      Val (s, Lines, ValError);
      if (ValError <> 0) or (Lines < 0) then
        begin
          ReadGPDError ('invalid entry');
          Exit
        end;
      if not ReadMD5 ('m ', MD5) then Exit;
      if not ReadMD5 ('m ', MD5Units[True]) then Exit;
      if not ReadMD5 ('m ', MD5Units[False]) then Exit;
      if not ReadLine ('M ', ModuleName) then Exit;
      PExportedInterfaces := @ExportedInterfaces;
      PIncludes := @Includes;
      PLinkedLibs := @LinkedLibs;
      PLinkedObjects := @LinkedObjects;
      PLinkedFiles := @LinkedFiles;
      PImplementationUses := @UsedUnits[False];
      PInterfaceUses := @UsedUnits[True];
      while not EOF (f) do
        begin
          NextLine;
          if (Length (Line) < 2) or (Line[2] <> ' ') then
            begin
              ReadGPDError ('invalid entry');
              Exit
            end;
          case Line[1] of
            'e': if not AddStringList (PascalFlag, PExportedInterfaces) then Exit;
            'i': if not AddStringList (True, PIncludes) then Exit;
            'L': if not AddStringList (PascalFlag, PLinkedLibs) then Exit;
            'o': if not AddStringList (PascalFlag, PLinkedObjects) then Exit;
            'l': if not AddStringList (PascalFlag, PLinkedFiles) then Exit;
            'U': if not AddUnitList (PInterfaceUses) then Exit;
            'u': if not AddUnitList (PImplementationUses) then Exit;
            else
              ReadGPDError ('invalid entry');
              Exit
          end
        end;
      PExportedInterfaces^ := nil;
      PIncludes^ := nil;
      PLinkedLibs^ := nil;
      PLinkedObjects^ := nil;
      PLinkedFiles^ := nil;
      PInterfaceUses^ := nil;
      PImplementationUses^ := nil
    end;
  Close (f);
  if Preliminary then
    ReadGPD := gd_Preliminary
  else
    ReadGPD := gd_Final
end;

procedure WriteGPD (const Dep: TDep; Preliminary, SetTime: Boolean);
var
  f: Text;
  ModificationTime: UnixTimeType;

  procedure WriteStringList (const Prefix: String; p: PStringList);
  begin
    while p <> nil do
      begin
        WriteLn (f, Prefix, p^.s);
        p := p^.Next
      end
  end;

  procedure WriteUnitList (const Prefix: String; p: PUnitList);
  begin
    while p <> nil do
      begin
        WriteLn (f, Prefix, p^.InterfaceName, ' ', p^.FileName, ' ', MD5Str (p^.MD5));
        p := p^.Next
      end
  end;

begin
  if SetTime then
    begin
      ModificationTime := GetModificationTime (Dep.ObjectName);
      if (ModificationTime < 0) then
        Error ('internal error: `' + Dep.ObjectName + ''' not found')
    end
  else
    ModificationTime := 0;
  if MkDirFlag then CreateDir (DirFromPath (Dep.GPDName));
  AssignBinary (f, Dep.GPDName);
  Rewrite (f);
  with Dep do
    begin
      WriteLn (f, GPDHeader[Preliminary]);
      WriteLn (f, 'T ', Ord (Kind));
      WriteLn (f, 't ', Platform);
      WriteLn (f, 'V ', Version);
      WriteLn (f, 'P ', Options);
      if Kind = fk_Program then WriteLn (f, 'p ', LinkOptionsStr);
      if not Preliminary then
        begin
          WriteLn (f, 's ', SrcName);
          WriteLn (f, 'O ', ObjectName)
        end;
      WriteLn (f, 'g ', GPIName);
      WriteLn (f, 'n ', Lines);
      WriteLn (f, 'm ', MD5Str (MD5));
      WriteLn (f, 'm ', MD5Str (MD5Units[True]));
      WriteLn (f, 'm ', MD5Str (MD5Units[False]));
      WriteLn (f, 'M ', ModuleName);
      WriteStringList ('e ', ExportedInterfaces);
      WriteStringList ('i ', Includes);
      WriteStringList ('L ', LinkedLibs);
      WriteStringList ('o ', LinkedObjects);
      WriteStringList ('l ', LinkedFiles);
      WriteUnitList ('U ', UsedUnits[True]);
      WriteUnitList ('u ', UsedUnits[False])
    end;
  Close (f);
  if SetTime then SetFileTime (f, ModificationTime, ModificationTime)
end;

{ Main functionality }

{ Unfortunately, we need 4 sets of options for each compiler:
  - only version options when getting the version, platform paths,
    etc., because GPC has problems handling other options there
  - all but `-v' for `gpc -E ...' etc., so the verbose output will
    not mess up the information gathered via standard error
  - all but linker options for normal compilations and linking
  - all options for linking }
procedure StartPipe (PascalFlag: Boolean; OptionSelection: TOptionSelection;
                     const AdditionalParameters: array [m .. n: Integer] of PString;
                     var OutputFile, StdErrFile: AnyFile; var Status: Integer;
                     var CmdLine: String);  { CmdLine may be Null }
var
  Process: PPipeProcess;
  Parameters: PPStrings;
  OptionCount, i, l, e: Integer;
  s: PString;
  p: POptionList;
begin
  if PascalFlag and (NeededOptions <> nil) then
    e := NeededOptions^.Count
  else
    e := 0;
  OptionCount := Compilers[PascalFlag].OptionCount[OptionSelection];
  New (Parameters, e + OptionCount + n - m + 1);
  for i := 1 to e do
    Parameters^[i] := NeededOptions^[i];
  i := e;
  p := Compilers[PascalFlag].Options[OptionSelection];
  while p <> nil do
    begin
      Inc (i);
      Parameters^[i] := @p^.Option;
      p := p^.Next
    end;
  Assert (i = e + OptionCount);
  for i := 1 to n - m + 1 do
    Parameters^[e + OptionCount + i] := NewString (AdditionalParameters[i + m - 1]^);
  if (Verbosity >= 2) or (@CmdLine <> nil) then
    begin
      l := Length (Compilers[PascalFlag].ProgName);
      for i := 1 to Parameters^.Count do
        Inc (l, Length (Parameters^[i]^) + 1);
      New (s, Max (1, l));
      s^ := Compilers[PascalFlag].ProgName;
      for i := 1 to Parameters^.Count do
        s^ := s^ + ' ' + Parameters^[i]^;
      if @CmdLine <> nil then CmdLine := s^;
      Message (2, s^);
      Dispose (s)
    end;
  ProgressMessageStartSubprocess;
  Pipe (Null, OutputFile, StdErrFile, Compilers[PascalFlag].ProgName, Parameters^, GetCEnvironment, Process, nil);
  Status := 0;
  Process^.Status := @Status;
  for i := e + OptionCount + 1 to Parameters^.Count do
    Dispose (Parameters^[i]);
  Dispose (Parameters);
  if (@OutputFile = nil) and (@StdErrFile = nil) then Discard (WaitPipeProcess (Process))
end;

procedure InitCompiler (PascalFlag: Boolean);
const
  CPascal: array [0 .. 2] of String (8) = ('C', 'Pascal', 'Pascal/C');
var
  Parameters1: array [1 .. 1] of PString = (@'-dumpmachine');
  Parameters2: array [1 .. 1] of PString = (@'-dumpversion');
  Both: Boolean;
  OutputFile: Text;
  Status: Integer;
  CmdLine, s: TString;
begin
  if Compilers[PascalFlag].Initialized then Exit;
  Both := Compilers[False].ProgName = Compilers[True].ProgName;
  ProgressMessage ('Checking ' + CPascal[Ord (PascalFlag and not Both) + 2 * Ord (Both)] + ' compiler');
  StartPipe (PascalFlag, co_OnlyVersion, Parameters1, OutputFile, Null, Status, CmdLine);
  ReadLn (OutputFile, s);
  Compilers[PascalFlag].Platform := s;
  if Both then Compilers[not PascalFlag].Platform := s;
  if not EOF (OutputFile) then
    Error ('internal problem: unexpected additional output of `' + CmdLine + '''');
  Close (OutputFile);
  if Status <> 0 then Quit (Status);
  StartPipe (PascalFlag, co_OnlyVersion, Parameters2, OutputFile, Null, Status, CmdLine);
  ReadLn (OutputFile, s);
  Compilers[PascalFlag].Version := s;
  if Both then Compilers[not PascalFlag].Version := s;
  if not EOF (OutputFile) then
    Error ('internal problem: unexpected additional output of `' + CmdLine + '''');
  Close (OutputFile);
  if Status <> 0 then Quit (Status);
  Compilers[PascalFlag].Initialized := True;
  if Both then
    Compilers[not PascalFlag].Initialized := True
  else if CheckPlatform and Compilers[not PascalFlag].Initialized
          and (Compilers[True].Platform <> Compilers[False].Platform) then
    Error ('Pascal and C compilers specified seem to be for different platforms' + NewLine +
           '(`' + Compilers[True].Platform + ''' and `' + Compilers[False].Platform + ''')');
  ProgressMessage ('')
end;

function UnitNameMatch (const Dep: TDep; const FileName, InterfaceName: String): Boolean;
begin
  if FileName <> '' then
    UnitNameMatch := (Dep.SrcBaseNameExt = FileName) or (Dep.SrcBaseName = FileName)
  else
    UnitNameMatch := Dep.SrcBaseName = InterfaceName
end;

function NewUsesElement (const InterfaceName, FileName: String; Next: PUnitList) = p: PUnitList;
begin
  New (p);
  p^.InterfaceName := LoCaseStr (InterfaceName);
  p^.FileName := FileName;
  MD5Clear (p^.MD5);
  p^.Dep := nil;
  p^.Next := Next
end;

procedure GetDep (var Dep: TDep);

  function ReadBufferAndIncludes (var OutputFile: file; var StdErrFile: Text; var Buffer: PBytes;
                                  var BufCount, Lines: Integer; var Includes: PStringList) = OK: Boolean;
  var
    NewBuffer: PBytes;
    Count, BytesRead, Select, Last, c, i, j: Integer;
    NewStyle: Boolean;
    ppi: PPStringList;
    s: TString;
    Files: array [1 .. 2] of PAnyFile;
  begin
    OK := True;
    NewStyle := False;  { unfortunately different gcc versions produce slightly incompatible output }
    Count := 0;
    New (Buffer, $40000);
    ppi := @Includes;
    Files[1] := @OutputFile;
    Files[2] := @StdErrFile;
    while (Files[1] <> nil) or (Files[2] <> nil) do
      begin
        Select := IOSelectRead (Files, -1);
        if Select <= 0 then Select := 1 + Ord (Files[1] = nil);
        case Select of
          1: begin
               BlockRead (OutputFile, Buffer^[Count], Buffer^.Size - Count, BytesRead);
               if BytesRead = 0 then
                 Files[Select] := nil
               else
                 begin
                   Inc (Count, BytesRead);
                   if Count >= Buffer^.Size then
                     begin
                       New (NewBuffer, Buffer^.Size * 2);
                       Move (Buffer^[0], NewBuffer^[0], Count);
                       Dispose (Buffer);
                       Buffer := NewBuffer
                     end
                 end
             end;
          2: if EOF (StdErrFile) then
               Files[Select] := nil
             else
               begin
                 ReadLn (StdErrFile, s);
                 if not NewStyle and (s <> '') and (s[1] = '.') then
                   begin
                     i := 1;
                     while (i <= Length (s)) and (s[i] = '.') do Inc (i);
                     if (i < Length (s)) and (s[i] = ' ') then
                       begin
                         NewStyle := True;
                         Message (4, 'new style GCC `-H'' output detected')
                       end
                   end;
                 if (Pos (': ', s) <> 0)
                    or IsPrefix ('In file included from', s)
                    or IsPrefix ('                 from', s) then
                   begin
                     WriteLn (StdErr, s);
                     OK := False
                   end
                 else if (s <> '') and (not NewStyle or (s[1] = '.')) then
                   begin
                     i := 1;
                     while (i <= Length (s)) and (s[i] = '.') do Inc (i);
                     if (i <= Length (s)) and (s[i] = ' ') then Delete (s, 1, i);
                     New (ppi^);
                     ppi^^.s := FExpand (Slash2OSDirSeparator (TrimBothStr (s)));
                     ppi := @ppi^^.Next
                   end
               end;
        end
      end;
    Close (OutputFile);
    ppi^ := nil;
    Close (StdErrFile);
    Lines := 0;
    j := 0;
    Last := 0;
    for i := 0 to Count - 1 do
      begin
        c := Buffer^[i];
        if ((Last = 13) and (c = 10)) or ((Last = 10) and (c = 13)) then
          Last := 0
        else
          begin
            Last := c;
            if c in [13, 10] then
              begin
                c := Ord (NewLine);
                Inc (Lines)
              end;
            Buffer^[j] := c;
            Inc (j)
          end
      end;
    Buffer^[j] := 0;
    BufCount := j
  end;

  procedure ParsePascal (var SrcName: String; var Dep: TDep);
  var
    Parameters: array [1 .. 3] of PString = (@'-E', @'-H', @SrcName);
    OutputFile: file;
    StdErrFile: Text;
    Buffer: PBytes;
    FileName, FullName: TString;
    ModuleName: CString;
    ExportedInterfaces, InterfaceUses, ImplementationUses, LinkedFiles, ResourceFiles, pt: PCStringList;
    Status, BufCount: Integer;
    ImplementationPos: CInteger;
    ppe, ppl, ppo, ppf: PPStringList;

    function CStringList2UnitList (pcs: PCStringList) = p: PUnitList;
    var
      t: PCStringList;
      pp: PPUnitList;
    begin
      pp := @p;
      while pcs <> nil do
        begin
          pp^ := NewUsesElement (CString2String (pcs^.Name1), CString2String (pcs^.Name2), nil);
          pp := @pp^^.Next;
          t := pcs;
          pcs := pcs^.Next;
          Dispose (t^.Name1);
          Dispose (t^.Name2);
          Dispose (t)
        end;
      pp^ := nil
    end;

  begin
    StartPipe (True, co_NoVerbose, Parameters, OutputFile, StdErrFile, Status, Null);
    if not ReadBufferAndIncludes (OutputFile, StdErrFile, Buffer, BufCount, Dep.Lines, Dep.Includes) and (Status = 0) then
      Status := 1;
    if Status <> 0 then Quit (Status);
    ExportedInterfaces := nil;
    InterfaceUses := nil;
    ImplementationUses := nil;
    LinkedFiles := nil;
    ResourceFiles := nil;
    ImplementationPos := 0;
    if BufCount = 0 then
      Dep.Kind := fk_None
    else
      Dep.Kind := ParseFile (SrcName, Buffer^[0], BufCount, ImplementationPos, ModuleName, ExportedInterfaces, InterfaceUses, ImplementationUses, LinkedFiles, ResourceFiles);
    MD5Buffer (Buffer^[0], ImplementationPos, Dep.MD5);
    Dispose (Buffer);
    if Dep.Kind = fk_None then
      begin
        if not NoWarnings then Warning (SrcName + ': missing program header');
        Dep.Kind := fk_Program
      end;
    Dep.ModuleName := CString2String (ModuleName);
    Dep.UsedUnits[False] := CStringList2UnitList (ImplementationUses);
    Dep.UsedUnits[True] := CStringList2UnitList (InterfaceUses);
    ppe := @Dep.ExportedInterfaces;
    while ExportedInterfaces <> nil do
      begin
        New (ppe^);
        ppe^^.s := CString2String (ExportedInterfaces^.Name1);
        ppe := @ppe^^.Next;
        pt := ExportedInterfaces;
        ExportedInterfaces := ExportedInterfaces^.Next;
        Dispose (pt^.Name1);
        Dispose (pt^.Name2);
        Dispose (pt)
      end;
    ppe^ := nil;
    ppl := @Dep.LinkedLibs;
    ppo := @Dep.LinkedObjects;
    ppf := @Dep.LinkedFiles;
    while LinkedFiles <> nil do
      begin
        FileName := CString2String (LinkedFiles^.Name1);
        case Ext2Language (ExtFromPath (FileName)) of
          fk_Object: begin
                       FullName := FSearch (FileName, ObjectPath);
                       if FullName <> '' then FullName := FExpand (FullName);
                       if FullName = '' then
                         Error ('object file `' + FileName + ''' not found');
                       New (ppo^);
                       ppo^^.s := FullName;
                       ppo := @ppo^^.Next
                     end;
          fk_None  : begin
                       New (ppl^);
                       ppl^^.s := '-l' + FileName;
                       ppl := @ppl^^.Next
                     end;
          else       New (ppf^);
                     ppf^^.s := FileName;
                     ppf := @ppf^^.Next
        end;
        pt := LinkedFiles;
        LinkedFiles := LinkedFiles^.Next;
        Dispose (pt^.Name1);
        Dispose (pt^.Name2);
        Dispose (pt)
      end;
    while ResourceFiles <> nil do
      begin
        FileName := CString2String (ResourceFiles^.Name1);
        FullName := FSearch (FileName, ObjectPath);
        if FullName <> '' then FullName := FExpand (FullName);
        if FullName = '' then
          Error ('resource file `' + FileName + ''' not found');
        New (ppl^);
        ppl^^.s := FullName + '.resource';
        ppl := @ppl^^.Next;
        pt := ResourceFiles;
        ResourceFiles := ResourceFiles^.Next;
        Dispose (pt^.Name1);
        Dispose (pt^.Name2);
        Dispose (pt)
      end;
    ppl^ := nil;
    ppo^ := nil;
    ppf^ := nil
  end;

  procedure ParseC (var SrcName: String; var Dep: TDep);
  var
    Parameters: array [1 .. 3] of PString = (@'-E', @'-H', @SrcName);
    StdErrFile: Text;
    OutputFile: file;
    Buffer: PBytes;
    Status, BufCount: Integer;
  begin
    StartPipe (False, co_NoVerbose, Parameters, OutputFile, StdErrFile, Status, Null);
    if not ReadBufferAndIncludes (OutputFile, StdErrFile, Buffer, BufCount, Dep.Lines, Dep.Includes) and (Status = 0) then
      Status := 1;
    if Status <> 0 then Quit (Status);
    Dispose (Buffer);
    Dep.ExportedInterfaces := nil;
    Dep.LinkedLibs := nil;
    Dep.LinkedObjects := nil;
    Dep.LinkedFiles := nil;
    Dep.ModuleName := '';
    Dep.UsedUnits[False] := nil;
    Dep.UsedUnits[True] := nil
  end;

  function SourceNewerThan (Time1, Time2: UnixTimeType; const Msg: String): Boolean;
  var
    NewestSourceTime, IncludeTime: UnixTimeType;
    Include: PStringList;
  begin
    SourceNewerThan := False;
    NewestSourceTime := GetModificationTime (Dep.SrcName);
    Include := Dep.Includes;
    while (Include <> nil) and (NewestSourceTime >= 0) do
      begin
        IncludeTime := GetModificationTime (Include^.s);
        if (IncludeTime < 0) or (IncludeTime > NewestSourceTime) then
          NewestSourceTime := IncludeTime;
        Include := Include^.Next
      end;
    if NewestSourceTime < 0 then
      begin
        Message (3, Dep.SrcName + ': ' + Msg + ' because not all previous source files exist');
        SourceNewerThan := True
      end
    else if (NewestSourceTime > Time1) or ((Time2 >= 0) and (NewestSourceTime > Time2)) then
      begin
        Message (3, Dep.SrcName + ': ' + Msg + ' because source has changed');
        SourceNewerThan := True
      end
  end;

var
  SavedSrcName, SavedSrcBaseNameExt, ObjectName, GPIName: TString;
  SrcExists, PascalFlag: Boolean;
  GPD: TGPD;
  Kind: TFileKind;
  GPDTime, GPITime, ObjectTime: UnixTimeType;
  pu, pua: PUnitList;

begin
  Kind := Dep.Kind;
  case Kind of
    fk_Program,
    fk_Unit,
    fk_Module: PascalFlag := True;
    fk_CFile:  PascalFlag := False;
    else       Error ('`' + Dep.SrcName + ''' is neither Pascal nor C')
  end;
  InitCompiler (PascalFlag);
  Dep.GPDName := Dep.ObjectBasePath + '.gpd';
  if PascalFlag then
    GPIName := Dep.GPIBasePath + '.gpi'
  else
    GPIName := '';
  ObjectName := Dep.ObjectBasePath + '.o';
  Dep.Recompile := rc_None;
  SrcExists := FileExists (Dep.SrcName);
  GPDTime := GetModificationTime (Dep.GPDName);
  ObjectTime := GetModificationTime (ObjectName);
  SavedSrcName := Dep.SrcName;
  SavedSrcBaseNameExt := Dep.SrcBaseNameExt;
  GPD := ReadGPD (Dep);
  if GPD = gd_None then
    GPITime := -1
  else
    GPITime := GetModificationTime (Dep.GPIName);
  if (ObjectTime < 0) or (GPDTime <> ObjectTime) or (GPD <> gd_Final)
     or ((Dep.GPIName <> '') and (GPITime < 0)) then
    if not SrcExists then
      ErrorFile (SavedSrcBaseNameExt, 'file not found')
    else
      begin
        Dep.SrcName := SavedSrcName;
        Dep.Recompile := rc_AlwaysButGPIOK
      end
  else if not SrcExists then
    begin
      if NameExtFromPath (Dep.SrcName) <> SavedSrcBaseNameExt then
        ErrorFile (SavedSrcBaseNameExt, 'file not found');
      Dep.SrcName := Dep.SrcBaseNameExt;
      if Cleaning = 0 then
        if AutoBuild then
          Message (0, Dep.SrcName + ': cannot recompile because source not found in spite of `--autobuild''')
        else
          Message (1, Dep.SrcName + ': cannot recompile because source not found');
      Dep.ObjectName := ObjectName;
      if Dep.Kind <> fk_Program then Dep.GPIName := GPIName;
      Dep.Recompile := rc_Impossible
    end
  else if AutoBuild then
    begin
      Dep.SrcName := SavedSrcName;
      if Cleaning = 0 then
        Message (3, Dep.SrcName + ': recompiling because of `--autobuild''');
      Dep.Recompile := rc_Always
    end
  else if (Dep.SrcName <> SavedSrcName)
         or (Dep.ObjectName <> ObjectName)
         or ((Dep.GPIName <> GPIName) and ((Dep.Kind <> fk_Program) or (Dep.GPIName <> ''))) then
    begin
      Dep.SrcName := SavedSrcName;
      if Cleaning = 0 then
        Message (3, Dep.SrcName + ': recompiling because of file name mismatch');
      Dep.Recompile := rc_Always
    end
  else if (Cleaning = 0) and SourceNewerThan (ObjectTime, GPITime, 'recompiling') then
    Dep.Recompile := rc_SrcChanged;
  if (Dep.Recompile <> rc_Always) and (Cleaning = 0) then
    begin
      if CheckPlatform and (Dep.Platform <> Compilers[PascalFlag].Platform) then
        if Dep.Recompile = rc_Impossible then
          Message (0, Dep.SrcName + ': platform mismatch and cannot recompile')
        else
          begin
            if Dep.Recompile < rc_SrcChanged then
              Message (3, Dep.SrcName + ': recompiling because of platform mismatch');
            Dep.Recompile := rc_Always
          end;
      if CheckVersion and (Dep.Version <> Compilers[PascalFlag].Version) then
        if Dep.Recompile = rc_Impossible then
          Message (1, Dep.SrcName + ': compiler version mismatch and cannot recompile')
        else
          begin
            if Dep.Recompile < rc_SrcChanged then
              Message (3, Dep.SrcName + ': recompiling because of compiler version mismatch');
            Dep.Recompile := rc_Always
          end;
      if CheckOptions and (Dep.Options <> Compilers[PascalFlag].OptionStr) then
        if Dep.Recompile = rc_Impossible then
          Message (1, Dep.SrcName + ': options mismatch and cannot recompile')
        else
          begin
            if Dep.Recompile < rc_SrcChanged then
              Message (1, Dep.SrcName + ': recompiling because of options mismatch');
            Dep.Recompile := rc_Always
          end
    end;
  if Dep.Recompile >= rc_SrcChanged then
    begin
      Dep.ObjectName := ObjectName;
      if AutoBuild or (GPD = gd_None)
         or (CheckPlatform and (Dep.Platform <> Compilers[PascalFlag].Platform))
         or (CheckVersion and (Dep.Version <> Compilers[PascalFlag].Version))
         or (CheckOptions and (Dep.Options <> Compilers[PascalFlag].OptionStr))
         or ((Cleaning = 0) and SourceNewerThan (GPDTime, -1, 'rechecking dependencies')) then
        begin
          ProgressMessage ('Checking dependencies of ' + Dep.SrcName);
          Dep.Platform := Compilers[PascalFlag].Platform;
          Dep.Version := Compilers[PascalFlag].Version;
          Dep.Options := Compilers[PascalFlag].OptionStr;
          MD5Clear (Dep.MD5);
          Dep.MD5Units[False] := Dep.MD5;
          Dep.MD5Units[True] := Dep.MD5;
          if PascalFlag then
            ParsePascal (Dep.SrcName, Dep)
          else
            ParseC (Dep.SrcName, Dep);
          if Dep.Kind = fk_Program then GPIName := '';
          Dep.GPIName := GPIName;
          if not TestParserFlag then WriteGPD (Dep, True, False);
          ProgressMessage ('');
          GPITime := -1
        end
      else
        begin
          if Dep.Kind = fk_Program then GPIName := '';
          Dep.GPIName := GPIName
        end;
      if Dep.Recompile = rc_AlwaysButGPIOK then
        if (GPITime < 0) or SourceNewerThan (GPITime, -1, 'recompiling interface') then
          begin
            Dep.Recompile := rc_Always;
            if Cleaning = 0 then
              Message (3, Dep.SrcName + ': compiling because target files not found or not up to date')
          end
        else if Cleaning = 0 then
          Message (3, Dep.SrcName + ': compiling implementation because target files not found or not up to date');
    end;
  if NormalizedKind[Kind] <> NormalizedKind[Dep.Kind] then
    if (Kind = fk_Program) and (NormalizedKind[Dep.Kind] = fk_Unit) then
      LinkFlag := False
    else if (Dep.Kind = fk_Program) and (NormalizedKind[Kind] = fk_Unit) then
      ErrorFile (Dep.SrcName, 'used as a unit, but is a program')
    else
      ErrorFile (Dep.SrcName, 'internal error: language mismatch');
  if Kind = fk_CFile then
    Dep.PBLines := Dep.Lines div CLinesDiv
  else
    Dep.PBLines := Dep.Lines;
  if not PascalFlag then Dep.ApplyAutoUses := False;
  if Dep.ApplyAutoUses then
    begin
      pua := AutoUsedUnits;
      while pua <> nil do
        begin
          { Don't add the unit again if it was already stored in the GPD file }
          pu := Dep.UsedUnits[True];
          while (pu <> nil) and (pu^.InterfaceName <> pua^.InterfaceName) do pu := pu^.Next;
          if pu = nil then
            Dep.UsedUnits[True] := NewUsesElement (pua^.InterfaceName, pua^.FileName, Dep.UsedUnits[True]);
          pua := pua^.Next
        end
    end
end;

{ Within this function, Deps is the list of all modules (program,
  units, modules, C files) discovered so far. DepToDo is a pointer
  within this list to the one currently being handled. A (template
  for a) module is added to the end of Deps as soon as a reference
  to it is found. So, the list elements of Deps before DepsToDo
  contain modules whose dependencies have been checked, while the
  ones after DepsToDo contain templates for the other ones. }
function CollectDeps (const MainSrc: String) = Deps: PDepList;
var
  DepToDo, CurrentDep: PDepList;
  pp: PPDepList;
  pu: PUnitList;
  ps: PStringList;
  DoInterface, DoApplyAutoUses: Boolean;

  function NewDep (const aSrcName, aInterfaceName: TString; aKind: TFileKind; aMain: Boolean) = p: PDepList;
  var
    Dir, BaseName, Ext, DestDir, Suffixes: TString;
    i: Integer;
    ExtKind: TFileKind;
  begin
    New (p);
    New (p^.Dep);
    with p^.Dep^ do
      begin
        UsedBy[False] := nil;
        UsedBy[True] := nil;
        LinkOptionsMismatch := False;
        FSplit (aSrcName, Dir, BaseName, Ext);
        SrcName := aSrcName;
        SrcBaseName := BaseName;
        SrcBaseNameExt := BaseName + Ext;
        InterfaceName := aInterfaceName;
        ExtKind := Ext2Language (Ext);
        ApplyAutoUses := DoApplyAutoUses;
        if (aKind = fk_CFile) and (ExtKind <> fk_CFile) then
          aKind := fk_Unit;  { accept Pascal files linked with linker directives }
        Kind := aKind;
        Main := aMain;
        if LanguageNormalizedKind[Kind] <> LanguageNormalizedKind[ExtKind] then
          if Kind = fk_Program then
            begin
              Suffixes := '';
              for i := Low (LanguageSuffixes[fk_Program]) to High (LanguageSuffixes[fk_Program]) do
                if LanguageSuffixes[fk_Program, i] <> '' then
                  Suffixes := Suffixes + ', `' + LanguageSuffixes[fk_Program, i] + '''';
              Delete (Suffixes, 1, 1);
              ErrorFile (SrcName, 'only a Pascal file (suffixes ' + Suffixes + ')' + NewLine + 'allowed on the command line')
            end
          else
            ErrorFile (SrcName, 'internal error: language file name suffix mismatch');
        if Kind in [fk_CFile, fk_Object] then
          DestDir := ObjectDestinationPath
        else
          DestDir := UnitDestinationPath;
        if DestDir = '' then DestDir := Dir;
        ObjectBasePath := DestDir + BaseName;
        GPIBasePath := DestDir + aInterfaceName
      end;
    p^.Next := nil
  end;

  { Search module and, if not found, create a template and add it at the end of Deps }
  function FindDep (const Path, FileName, InterfaceName: TString; Kind: TFileKind): PDepList;
  var
    pp: PPDepList;
    i: Integer;
    SrcName, Base: TString;
  begin
    pp := @Deps;
    while (pp^ <> nil) and
      not ((NormalizedKind[pp^^.Dep^.Kind] = NormalizedKind[Kind]) and
           UnitNameMatch (pp^^.Dep^, FileName, InterfaceName)) do
      pp := @pp^^.Next;
    if pp^ = nil then
      begin
        if (FileName <> '') and (ExtFromPath (FileName) <> '') then
          begin
            SrcName := FSearch (FileName, Path);
            if SrcName <> '' then SrcName := FExpand (SrcName);
            if SrcName = '' then SrcName := FileName
          end
        else
          begin
            if FileName <> '' then
              Base := FileName
            else
              Base := InterfaceName;
            for i := Low (LanguageSuffixes[Kind]) to High (LanguageSuffixes[Kind]) do
              if LanguageSuffixes[Kind, i] <> '' then
                begin
                  SrcName := FSearch (Base + LanguageSuffixes[Kind, i], Path);
                  if SrcName <> '' then Break
                end;
            if SrcName <> '' then SrcName := FExpand (SrcName);
            if SrcName = '' then
              SrcName := Base + LanguageSuffixes[Kind, Low (LanguageSuffixes[Kind])]
          end;
        pp^ := NewDep (SrcName, InterfaceName, Kind, False)
      end;
    FindDep := pp^
  end;

  procedure AddDepList (var List: PDepList; Dep: PDep);
  var p: PDepList;
  begin
    New (p);
    p^.Dep := Dep;
    p^.Next := List;
    List := p
  end;

begin
  Deps := nil;
  { Avoid creating cyclic or self dependencies due to `--uses' options.
    To facilitate this, process auto-imports first, disabling auto-imports for them. }
  DoApplyAutoUses := False;
  pu := AutoUsedUnits;
  while pu <> nil do
    begin
      Discard (FindDep (UnitPath, pu^.FileName, pu^.InterfaceName, fk_Unit));
      pu := pu^.Next
    end;
  DepToDo := Deps;
  while (DepToDo <> nil) or not DoApplyAutoUses do
    if DepToDo = nil then
      begin
        { Now we've handled all auto-imports and their dependencies. Proceed with
          the actually given sources, and activate auto-imports from now on. }
        DoApplyAutoUses := True;
        pp := @Deps;
        while pp^ <> nil do pp := @pp^^.Next;
        pp^ := NewDep (MainSrc, '', fk_Program, True);
        DepToDo := pp^
      end
    else
      begin
        GetDep (DepToDo^.Dep^);
        for DoInterface := False to True do
          begin
            pu := DepToDo^.Dep^.UsedUnits[DoInterface];
            while pu <> nil do
              begin
                CurrentDep := FindDep (UnitPath, pu^.FileName, pu^.InterfaceName, fk_Unit);
                AddDepList (CurrentDep^.Dep^.UsedBy[DoInterface], DepToDo^.Dep);
                pu^.Dep := CurrentDep^.Dep;
                pu := pu^.Next
              end
          end;
        ps := DepToDo^.Dep^.LinkedFiles;
        while ps <> nil do
          begin
            { Just make sure the file is added to Deps. We don't need
              any dependency relations because files linked with
              linker directives can be compiled at any time. }
            Discard (FindDep (ObjectPath, ps^.s, '', fk_CFile));
            ps := ps^.Next
          end;
        DepToDo := DepToDo^.Next
      end
end;

{ Check if any used unit's interface has changed (by comparing the
  MD5), and if so, update the MD5 and recompile this module.

  Don't abort the checking prematurely when it's clear that the
  current module has to be recompiled because we have to update the
  MD5 for all units.

  Note: The information whether to recompile the current module is
  stored in a field of Dep (`Recompile') rather than a local
  variable, so if InterfaceOnly is True, the information will still
  be available when later compiling the same module without
  `--interface-only'.

  MD5Units is recursively composed from the MD5 values of all units
  used in the interface or implementation, respectively, directly or
  indirectly through other modules' interfaces, so it's responsible
  for recompiling the current module if a module used indirectly was
  changed. }
procedure CompileIfNecessary (var Dep: TDep; InterfaceOnly: Boolean);
const
  InterfaceOf: array [0 .. 2] of String (18) = ('', 'interface of ', 'implementation of ');
var
  UsedUnitChanged, DoInterface, DoAssemble, ImplementationOnly: Boolean;
  p: PUnitList;
  NewMD5Units: TMD5;
  Status: Integer;
  GPIDestinationPath, GPIDestinationPathAsUnitPath, CompileOption, ObjectName: TString;
  Parameters: array [1 .. 10] of PString =
    (@'--interface-only', @'--no-progress-bar',
     @GPIDestinationPath, @GPIDestinationPathAsUnitPath, @'--no-automake',
     @CompileOption, @'-o', @ObjectName, @Dep.SrcName, @AutoUsesOption);
begin
  { If recompiling anyway, set UsedUnitChanged to suppress further messages }
  UsedUnitChanged := Dep.Recompile in [rc_SrcChanged, rc_Always];
  Dep.InterfaceUsedUnitRecompiled := False;
  for DoInterface := InterfaceOnly to True do
    begin
      NewMD5Units := Dep.MD5;
      p := Dep.UsedUnits[DoInterface];
      while p <> nil do
        begin
          if not MD5Compare (p^.MD5, p^.Dep^.MD5) then
            begin
              p^.MD5 := p^.Dep^.MD5;
              if not UsedUnitChanged then
                Message (3, Dep.SrcName + ': recompiling because interface of `' + p^.Dep^.SrcBaseNameExt + ''' has changed');
              UsedUnitChanged := True
            end
{$if False}
{ I think that's too paranoid. If options have an influence, they
  will affect the GPI's MD5. OTOH, this check forces recompilation
  if a used interface is changed forth and back before this module
  is updated. (If necessary, we could handle this latter case using
  a new rc_... value, but perhaps we can just forget about this
  check.) -- Frank, 20040329 }
          else if ((p^.Dep^.Recompile = rc_Always) or (p^.Dep^.InterfaceUsedUnitRecompiled))
                  and (Dep.Recompile <> rc_Impossible) then
            begin
              { A dependent module was recompiled for reasons other
                than changed source, such as changed options, which
                might also influence this module, so we recompile it
                as well if at all possible. This check, however, is
                not fail-safe when gp runs are aborted between
                compilations or some modules are recompiled without
                using gp. }
              if not UsedUnitChanged then
                if p^.Dep^.Recompile = rc_Always then
                  Message (3, Dep.SrcName + ': recompiling because `' + p^.Dep^.SrcBaseNameExt + ''' was recompiled')
                else
                  Message (3, Dep.SrcName + ': recompiling because a unit used indirectly was recompiled');
              UsedUnitChanged := True;
              if DoInterface then Dep.InterfaceUsedUnitRecompiled := True
            end
{$endif} ;
          NewMD5Units := MD5Compose (NewMD5Units, p^.Dep^.MD5Units[True]);
          p := p^.Next
        end;
      if not MD5Compare (Dep.MD5Units[DoInterface], NewMD5Units) then
        begin
          Dep.MD5Units[DoInterface] := NewMD5Units;
          if not UsedUnitChanged then
            Message (3, Dep.SrcName + ': recompiling because an interface used indirectly has changed');
          UsedUnitChanged := True
        end
    end;
  if UsedUnitChanged then
    if Dep.Recompile = rc_Impossible then
      ErrorFile (Dep.SrcName, 'must recompile because an interface used has changed,' + NewLine +
                              'but source is not available')
    else
      Dep.Recompile := Max (Dep.Recompile, rc_SrcChanged);
  DoAssemble := InterfaceOnly or not Dep.Main or AssembleFlag;
  if not DoAssemble or
     ((Dep.Recompile >= rc_SrcChanged) and not (InterfaceOnly and (Dep.Recompile = rc_AlwaysButGPIOK))) then
    begin
      CompileOption := '-c';
      ObjectName := Dep.ObjectName;
      if not DoAssemble then
        begin
          CompileOption := '-S';
          if MainDestNameGiven then
            ObjectName := MainDestName
          else
            ObjectName := Dep.SrcBaseName + '.s'
        end;
      if ObjectName = '' then Error ('internal error: missing object file name');
      GPIDestinationPath := DirFromPath (Dep.ObjectName);
      CreateDir (GPIDestinationPath);
      GPIDestinationPathAsUnitPath := '--unit-path=' + GPIDestinationPath;
      Insert ('--gpi-destination-path=', GPIDestinationPath, 1);
      ImplementationOnly := not InterfaceOnly and (Dep.DepsProcessed = dp_Interface);
      ProgressMessage ('Compiling ' + InterfaceOf[Ord (InterfaceOnly) + 2 * Ord (ImplementationOnly)]
                       + Dep.SrcBaseNameExt + ' (' + Integer2String (Dep.Lines) + ')');
      if InterfaceOnly then
        ProgressBarSet (LinesDone / TotalLines)
      else
        ProgressBarAdd (LinesDone / TotalLines, Dep.PBLines / TotalLines, Dep.PBLines);
      if ImplementationOnly then Parameters[2] := @'--implementation-only';
      StartPipe (Dep.Kind <> fk_CFile, co_Compile,
        Parameters[6 - 3 * Ord (Dep.Kind <> fk_CFile) - 2 * Ord (InterfaceOnly) - Ord (ImplementationOnly)
                   .. High (Parameters) - Ord ((AutoUses = '') or not Dep.ApplyAutoUses)],
        Null, Null, Status, Null);
      if Status <> 0 then Quit (Status);
      if DoAssemble then WriteGPD (Dep, InterfaceOnly, not InterfaceOnly);
      ProgressMessage ('')
    end
  else if not InterfaceOnly then
    Message (4, Dep.SrcName + ': source and all interfaces used are unchanged, not recompiling');
  if not InterfaceOnly then Inc (LinesDone, Dep.PBLines)
end;

{ Check all dependencies and recompile whatever necessary.

  To compile things in the right order, even in the presence of
  cyclic unit dependencies, we do a dual topological sort (with
  egdes representing the "used by" relation, i.e. we have an edge
  from B to A if A uses B). "Dual" means we label the egdes as
  `interface' and `implementation', depending on where the
  `uses'/`import' declaration occurs. (C files and Pascal files
  linked with a linker directive are handled like `implementation'
  uses, regardless of where the directive occurs, since they have no
  effect on the compilation of the module using them.) When there
  are no nodes without incoming edges (i.e., dependencies on other
  modules), we choose one with only incoming `implementation' edges.
  If it needs to be compiled, we do it with `--interface-only'.
  Afterwards, we remove the node's outgoing edges, but we don't
  forget about the node yet. After its incoming edges have been
  removed (i.e., all modules it depends on have been processed), it
  will be compiled again, this time without `--interface--only'.

  To realize this, we use two ZeroIn lists, one (indexed with False)
  for nodes without any incoming edges, and one (indexed with True)
  for nodes without incoming `interface' edges. When choosing nodes
  to process, the first one gets strict priority.

  We keep track of the InDegrees concerning `interface' and
  `implementation' edges separately. This guarantees that there are
  exactly two times (including possibly the start of the algorithm)
  where one of a given node's InDegrees reaches 0.

  When the `interface' InDegree reaches 0, the node is added to the
  first ZeroIn list if the `implementation' InDegree is already 0,
  and to the second one otherwise. When the `implementation'
  InDegree reaches 0, the node is added to the first list if the
  `interface' InDegree is 0, otherwise nothing is done.

  Therefore, if the `implementation' InDegree reaches 0 before the
  `interface' InDegree does, the node will only ever appear in the
  first ZeroIn list (this should be the common case, without cyclic
  dependencies). If the `interface' InDegree is the first to reach
  0, the node is added to the first list after is has been added to
  the second list. By this time it either has already been removed
  again from the second list (this is when it was considered to be
  compiled with `--interface-only'), or it is still there. For the
  latter case, since we don't remove it from the second list
  immediately (which would be an extra O(n) unless we make the list
  doubly linked), we maintain the `DepsProcessed' flag which, if set
  to `dp_All', causes the node to be dropped without further
  processing when found in the second list. }
procedure ProcessDependencies (Deps: PDepList);
var
  p, p2, pt, Cycle: PDepList;
  ZeroIn: array [Boolean] of PDepList;
  pu: PUnitList;
  pd: PDep;
  InterfaceOnly, DoInterface: Boolean;
  TotalCount, ProcessedCount: Integer;
  CycleStr: TString;

  {$ifdef DEBUG}
  procedure ShowInDegree (const Dep: TDep); attribute (inline);
  begin
    Message (5, 'DEBUG: ' + Dep.SrcBaseNameExt + ': InDegree=('
                + Integer2String (Dep.InDegree[True]) + ','
                + Integer2String (Dep.InDegree[False]) + ')')
  end;
  {$endif}

  procedure CountInDegree (Deps: PDepList; CareAboutInterface: Boolean);
  var
    p, p2: PDepList;
    DoInterface: Boolean;
  begin
    p := Deps;
    while p <> nil do
      begin
        p^.Dep^.InDegree[False] := 0;
        p^.Dep^.InDegree[True] := 0;
        p := p^.Next
      end;
    p := Deps;
    while p <> nil do
      begin
        for DoInterface := False to True do
          begin
            p2 := p^.Dep^.UsedBy[DoInterface];
            while p2 <> nil do
              begin
                Inc (p2^.Dep^.InDegree[DoInterface and CareAboutInterface]);
                p2 := p2^.Next
              end
          end;
        p := p^.Next
      end
  end;

  procedure AddZeroIn (DoInterface: Boolean; p: PDepList);
  var pt: PDepList;
  begin
    New (pt);
    pt^.Dep := p^.Dep;
    pt^.Next := ZeroIn[DoInterface];
    ZeroIn[DoInterface] := pt
  end;

begin
  ZeroIn[False] := nil;
  ZeroIn[True] := nil;
  TotalCount := 0;
  ProcessedCount := 0;
  LinesDone := 0;
  TotalLines := 0;
  CountInDegree (Deps, True);
  p := Deps;
  while p <> nil do
    with p^.Dep^ do
      begin
        Inc (TotalLines, PBLines);
        Inc (TotalCount);
        DepsProcessed := dp_None;
        CyclePosition := nil;
        if InDegree[True] = 0 then AddZeroIn (InDegree[False] <> 0, p);
        {$ifdef DEBUG}
        ShowInDegree (p^.Dep^);
        {$endif}
        p := p^.Next
      end;
  while (ZeroIn[False] <> nil) or (ZeroIn[True] <> nil) do
    begin
      InterfaceOnly := ZeroIn[False] = nil;
      p := ZeroIn[InterfaceOnly];
      ZeroIn[InterfaceOnly] := p^.Next;
      if p^.Dep^.DepsProcessed <> dp_All then
        begin
          CompileIfNecessary (p^.Dep^, InterfaceOnly);
          if p^.Dep^.DepsProcessed = dp_None then
            for DoInterface := False to True do
              begin
                p2 := p^.Dep^.UsedBy[DoInterface];
                while p2 <> nil do
                  with p2^.Dep^ do
                    begin
                      Dec (InDegree[DoInterface]);
                      if InDegree[DoInterface] = 0 then
                        if InDegree[not DoInterface] = 0 then
                          AddZeroIn (False, p2)
                        else if DoInterface then
                          AddZeroIn (True, p2);
                      {$ifdef DEBUG}
                      ShowInDegree (p2^.Dep^);
                      {$endif}
                      p2 := p2^.Next
                    end
              end;
          if InterfaceOnly then
            p^.Dep^.DepsProcessed := dp_Interface
          else
            begin
              p^.Dep^.DepsProcessed := dp_All;
              Inc (ProcessedCount)
            end
        end;
      Dispose (p)
    end;
  if ProcessedCount <> TotalCount then
    begin
      p := Deps;
      while (p <> nil) and (p^.Dep^.DepsProcessed = dp_All) do p := p^.Next;
      if p = nil then
        Error ('internal error: cycle not found');
      pd := p^.Dep;
      Cycle := nil;
      repeat
        New (pt);
        pt^.Dep := pd;
        pt^.Next := Cycle;
        Cycle := pt;
        pd^.CyclePosition := Cycle;
        pu := pd^.UsedUnits[True];
        while (pu <> nil) and (pu^.Dep^.DepsProcessed = dp_All) do pu := pu^.Next;
        if pu = nil then
          Error ('internal error: complete cycle not found');
        if pd = pu^.Dep then
          ErrorFile (pd^.SrcName, 'dependent on itself');
        pd := pu^.Dep
      until pd^.CyclePosition <> nil;
      CycleStr := Cycle^.Dep^.SrcBaseNameExt;
      while Cycle <> pd^.CyclePosition do
        begin
          Cycle := Cycle^.Next;
          CycleStr := Cycle^.Dep^.SrcBaseNameExt + ' -> ' + CycleStr
        end;
      Error ('cyclic interface dependency: ' + CycleStr)
    end
  else
    ProgressBarSet (1)
end;

procedure TestParser (const MainSrc: String);
var
  TestDep: TDep;
  Dir, BaseName, Ext: TString;
begin
  with TestDep do
    begin
      UsedBy[False] := nil;
      UsedBy[True] := nil;
      Kind := fk_Program;
      FSplit (MainSrc, Dir, BaseName, Ext);
      SrcName := MainSrc;
      SrcBaseName := BaseName;
      SrcBaseNameExt := BaseName + Ext;
      InterfaceName := LoCaseStr (BaseName);
      ObjectBasePath := Dir + BaseName;
      GPIBasePath := ObjectBasePath;
      ApplyAutoUses := True;
      GetDep (TestDep);
      GPDName := '-'
    end;
  WriteGPD (TestDep, False, False)
end;

function GetUnitPath = Res: TString;
var
  Parameters: array [1 .. 1] of PString = (@'-print-file-name=units');
  OutputFile: Text;
  Status: Integer;
begin
  InitCompiler (True);
  ProgressMessage ('Getting default unit path');
  StartPipe (True, co_OnlyVersion, Parameters, OutputFile, Null, Status, Null);
  ReadLn (OutputFile, Res);
  Close (OutputFile);
  if Status <> 0 then Quit (Status);
  if not DirectoryExists (Res) then
    Message (0, 'installation problem: cannot get default unit path');
  ProgressMessage ('')
end;

function GetNeededOptions: PPStrings;
var
  Parameters: array [1 .. 4] of PString = (@'--print-needed-options', @'-x', @'Preprocessed-Pascal', nil);
  OutputFile, f: Text;
  Status: Integer;
  s, s2: TString;
begin
  s := GetTempFileName;
  Rewrite (f, s);
  Close (f);
  Parameters[4] := @s;
  InitCompiler (True);
  ProgressMessage ('Getting needed options');
  StartPipe (True, co_OnlyVersion, Parameters, OutputFile, OutputFile, Status, Null);
  s := '';
  while not EOF (OutputFile) do
    begin
      ReadLn (OutputFile, s2);
      s := s + ' ' + s2
    end;
  Close (OutputFile);
  Erase (f);
  if Status <> 1 { sic ! } then Quit (Status + 1);
  ProgressMessage ('');
  GetNeededOptions := ShellTokenizeString (s, Null)
end;

function GetLibgccPath (PascalFlag: Boolean) = Res: TString;
var
  Parameters: array [1 .. 1] of PString = (@'-print-libgcc-file-name');
  OutputFile: Text;
  Status: Integer;
begin
  InitCompiler (True);
  ProgressMessage ('Getting libgcc path from ' + Compilers[PascalFlag].ProgName);
  StartPipe (PascalFlag, co_OnlyVersion, Parameters, OutputFile, Null, Status, Null);
  ReadLn (OutputFile, Res);
  Close (OutputFile);
  if Status <> 0 then Quit (Status);
  ProgressMessage ('')
end;

procedure Link (Deps: PDepList);
var
  ParCount, ParIndex, Status: Integer;
  LinkOptionsMismatch: Boolean;
  LibgccPath: TString;
  p: PDepList;
  ps: PStringList;
  ExecutableTime, ObjectTime: UnixTimeType;

  function Count (p: PStringList) = Res: Integer;
  begin
    Res := 0;
    while p <> nil do
      begin
        Inc (Res);
        p := p^.Next
      end
  end;

  procedure CheckObjectTime (const s: String);
  var Time: UnixTimeType;
  begin
    Time := GetModificationTime (s);
    if Time < 0 then Error ('internal error: object file `' + s + ''' not found');
    ObjectTime := Max (ObjectTime, Time)
  end;

begin
  ExecutableTime := GetModificationTime (MainDestName);
  if ExecutableTime >= 0 then
    begin
      ObjectTime := -1;
      LinkOptionsMismatch := False;
      p := Deps;
      while p <> nil do
        begin
          if p^.Dep^.LinkOptionsMismatch then
            begin
              Assert (p^.Dep^.Kind = fk_Program);
              LinkOptionsMismatch := True;
              WriteGPD (p^.Dep^, False, True)
            end;
          CheckObjectTime (p^.Dep^.ObjectName);
          ps := p^.Dep^.LinkedObjects;
          while ps <> nil do
            begin
              CheckObjectTime (ps^.s);
              ps := ps^.Next
            end;
          p := p^.Next
        end;
      if not LinkOptionsMismatch and (ObjectTime >= 0) and (ObjectTime <= ExecutableTime) then
        begin
          Message (4, MainDestName + ': newer than all object files, not relinking');
          Exit
        end
    end;
  LibgccPath := '';
  if not FileExists (GetLibgccPath (True)) then
    begin
      { If GPC does not find libgcc.a, we have to use GCC's one }
      LibgccPath := GetLibgccPath (False);
      if not FileExists (LibgccPath) then
        begin
          Message (0, 'installation problem: could not find libgcc.a, linking will probably fail');
          LibgccPath := ''
        end
      else
        LibgccPath := DirFromPath (LibgccPath)
    end;
  ParCount := 3;
  if LibgccPath <> '' then Inc (ParCount, 2);
  p := Deps;
  while p <> nil do
    begin
      Inc (ParCount, 1 + Count (p^.Dep^.LinkedObjects) + Count (p^.Dep^.LinkedLibs));
      p := p^.Next
    end;
  var Parameters: array [1 .. ParCount] of PString;
  Parameters[1] := @'-o';
  Parameters[2] := @MainDestName;
  Parameters[3] := @'--no-automake';
  ParIndex := 3;
  if LibgccPath <> '' then
    begin
      Inc (ParIndex, 2);
      Parameters[ParIndex - 1] := @'-L';
      Parameters[ParIndex] := @LibgccPath
    end;
  p := Deps;
  while p <> nil do
    begin
      Inc (ParIndex);
      Parameters[ParIndex] := @p^.Dep^.ObjectName;
      ps := p^.Dep^.LinkedObjects;
      while ps <> nil do
        begin
          Inc (ParIndex);
          Parameters[ParIndex] := @ps^.s;
          ps := ps^.Next
        end;
      p := p^.Next
    end;
  p := Deps;
  while p <> nil do
    begin
      ps := p^.Dep^.LinkedLibs;
      while ps <> nil do
        begin
          Inc (ParIndex);
          Parameters[ParIndex] := @ps^.s;
          ps := ps^.Next
        end;
      p := p^.Next
    end;
  Assert (ParIndex = ParCount);
  if MkDirFlag then CreateDir (DirFromPath (MainDestName));
  ProgressMessage ('Linking');
  StartPipe (True, co_All, Parameters, Null, Null, Status, Null);
  if Status <> 0 then Quit (Status);
  ProgressMessage ('')
end;

procedure DoClean (Dep: PDepList; Cleaning: Integer);
var p: PStringList;

  function Remove (const FileName: String): Boolean; attribute (ignorable);
  var
    f: Text;
    b: BindingType;
  begin
    Remove := False;
    Assign (f, FileName);
    b := Binding (f);
    if b.Bound and b.Existing then
      begin
        Message (2, 'rm ' + FileName);
        Erase (f);
        Remove := True
      end
  end;

begin
  while Dep <> nil do
    with Dep^.Dep^ do
      begin
        Remove (GPDName);
        Remove (ObjectName);
        if ExportedInterfaces <> nil then
          begin
            p := ExportedInterfaces;
            while p <> nil do
              begin
                Remove (UnitDestinationPath + LoCaseStr (p^.s) + '.gpi');
                p := p^.Next
              end
          end
        else if not Remove (GPIName) and (Kind in [fk_Unit, fk_Module]) and IsSuffixCase ('.o', ObjectName) then
          Remove (Copy (ObjectName, 1, Length (ObjectName) - 2) + '.gpi');
        if (Kind = fk_Module) and (ModuleName <> '') then
          Remove (UnitDestinationPath + LoCaseStr (ModuleName) + '-all.gpi');
        Dep := Dep^.Next
      end;
  if (Cleaning >= 2) and AssembleFlag and LinkFlag then
    Remove (MainDestName)
end;

{ Find all cyclic dependencies, i.e. the strongly connected
  components of the dependency graph, without distinguishing between
  `interface' and `implementation' uses, using Tarjan's algorithm. }
procedure DoPrintCycles (Deps: PDepList);
var
  p: PDepList;
  StackTop: PDep;
  Dummy: TDep;
  LastNumber, TotalCount, CycleCount: Integer;

  procedure HandleNode (Dep: PDep);
  var
    DoInterface, First: Boolean;
    p: PUnitList;
    pd: PDep;
  begin
    Inc (LastNumber);
    Dep^.Number := LastNumber;
    Dep^.LowLink := LastNumber;
    Dep^.StackNext := StackTop;
    StackTop := Dep;
    for DoInterface := False to True do
      begin
        p := Dep^.UsedUnits[DoInterface];
        while p <> nil do
          begin
            if p^.Dep^.Number = 0 then
              begin
                HandleNode (p^.Dep);
                Dep^.LowLink := Min (Dep^.LowLink, p^.Dep^.LowLink)
              end
            else if p^.Dep^.StackNext <> nil then
              Dep^.LowLink := Min (Dep^.LowLink, p^.Dep^.Number);
            p := p^.Next
          end
      end;
    if Dep^.LowLink = Dep^.Number then
      begin
        First := True;
        while StackTop^.Number >= Dep^.Number do
          begin
            pd := StackTop;
            StackTop := pd^.StackNext;
            pd^.StackNext := nil;
            if First then
              begin
                if StackTop^.Number < Dep^.Number then Break;  { omit 1-element cycles }
                First := False;
                WriteLn ('Cycle:')
              end;
            WriteLn ('  ', pd^.SrcName);
            Inc (CycleCount)
          end
      end
  end;

begin
  CycleCount := 0;
  TotalCount := 0;
  p := Deps;
  while p <> nil do
    begin
      Inc (TotalCount);
      p^.Dep^.Number := 0;
      p^.Dep^.LowLink := 0;
      p^.Dep^.StackNext := nil;
      p := p^.Next
    end;
  LastNumber := 0;
  Dummy.Number := 0;
  Dummy.StackNext := nil;
  StackTop := @Dummy;
  HandleNode (Deps^.Dep);
  Assert (StackTop = @Dummy);
  WriteLn ('Total modules: ', TotalCount);
  WriteLn ('Modules in cycles: ', CycleCount)
end;

procedure DoPrint (Dep: PDepList);
var
  p: PDepList;
  ps: PStringList;
begin
  if PrintSources then
    begin
      p := Dep;
      while p <> nil do
        begin
          WriteLn (p^.Dep^.SrcName);
          ps := p^.Dep^.Includes;
          while ps <> nil do
            begin
              WriteLn (ps^.s);
              ps := ps^.Next
            end;
          p := p^.Next
        end
    end;
  if PrintObjects then
    begin
      p := Dep;
      while p <> nil do
        begin
          WriteLn (p^.Dep^.ObjectName);
          ps := p^.Dep^.LinkedObjects;
          while ps <> nil do
            begin
              WriteLn (ps^.s);
              ps := ps^.Next
            end;
          p := p^.Next
        end
    end;
  if PrintLibs then
    begin
      p := Dep;
      while p <> nil do
        begin
          ps := p^.Dep^.LinkedLibs;
          while ps <> nil do
            begin
              WriteLn (ps^.s);
              ps := ps^.Next
            end;
          p := p^.Next
        end
    end;
  if PrintExecutable then
    if AssembleFlag and LinkFlag then
      WriteLn (MainDestName)
    else
      Error ('not linking, no executable name to print');
  if PrintCycles then
    DoPrintCycles (Dep)
end;

procedure Usage; attribute (noreturn);
begin
  WriteLn ('Usage: ', ProgramName, ' [options] source-file
Options:
--autobuild                    Compile all units, modules and C files
--no-autobuild                 Compile files only when necessary (default)
--automake                     Equivalent to `--no-autobuild'' (obsolete)
--clean                        Remove object, GPI, GPD and executable files
-c                             Compile only, do not link (default if main
                               source file is a unit or module rather than a
                               program)
--mkdir                        Create the executable and unit and object
                               destination paths if necessary (default)
--no-mkdir                     Don''t create the executable and unit and
                               object destination paths
--progress-messages            Write progress messages
--no-progress-messages         Do not write progress messages
--progress-bar                 Write messages for a progress bar
--no-progress-bar              Do not write messages for a progress bar
-v                             Set GP''s verbosity to the maximum, and make
                               the compilers verbose
-vN                            Set GP''s verbosity level to N (default: 0),
                               do not change the compilers'' verbosity
--unit-path=DIR[' + PathSeparator + 'DIR...]       Directories where to look for unit/module
                               sources (can be given multiple times)
--no-unit-path                 Forget about given `--unit-path''s
--object-path=DIR[' + PathSeparator + 'DIR...]     Directories where to look for additional object
                               files and source files in other languages in
                               addition to the `--unit-path''s given
--no-object-path               Forget about given `--object-path''s
--disable-default-paths        Do not add a default path to the unit and
                               object path
--include-path=DIR[' + PathSeparator + 'DIR...]    Directories where to look for include files
                               (same as `-I DIR1 -I DIR2 ...'', can be given
                               multiple times)
--library-path=DIR[' + PathSeparator + 'DIR...]    Directories where to look for libraries
                               (same as `-L DIR1 -L DIR2 ...'', can be given
                               multiple times)
--unit-destination-path=DIR    Path where to create object, GPI and GPD files
--no-unit-destination-path     Create object, GPI and GPD files in the current
                               directory (default)
--object-destination-path=DIR  Path where to create additional object and GPD
                               files (default: the path given with
                               `--unit-destination-path'')
--no-object-destination-path   Create additional object and GPD files in the
                               current directory
--executable-file-name=FILE    Base file name for the executable (default:
                               derived from the main program''s file name)
--executable-file-name         Ignored. Obsolete.
--executable-path=DIR          Directory in which to create the executable
--no-executable-path           Create the executable in the directory
                               where the main source is (default)
-o FILE                        Complete file name under which to create the
                               executable (overrides `--executable-path'' and
                               `--executable-file-name'')
--Pascal=OPTION                Use OPTION only for Pascal (where suitable)
--pascal=OPTION                Use OPTION only for Pascal (where suitable)
--c=OPTION                     Use OPTION only for C (where suitable)
PC=PATH                        Pascal compiler to use (default: `gpc'' unless
                               overridden by the environment variable `PC'')
CC=PATH                        C compiler to use (default: same as Pascal
                               compiler if set, otherwise `gcc'', unless
                               overridden by the environment variable `CC'')
--uses=UNIT[(FILE)][,...]      Add an implicit `uses'' clause
--no-uses                      Forget about all given `--uses'' options
--check-platform               Recompile if the target platform does not match
                               the one used for the previous compilation
                               (default)
--no-check-platform            Don''t compare the target platform
--check-version                Recompile if the compiler version does not
                               match the one used for the previous compilation
                               (default)
--no-check-version             Don''t compare the compiler version
--check-options                Recompile if the compiler options do not match
                               the ones used for the previous compilation
                               (default)
--no-check-options             Don''t compare the compiler options
--print-only                   Only do the print options, don''t compile
--print-platform               Print the target platform of the compiler used
--print-cycles                 Print all cyclic dependencies
--print-sources                Print the names of all source and include files
--print-objects                Print the names of all object files
--print-libs                   Print the names of all libraries used, prefixed
                               with `-l'' each, and resources, suffixed with
                               `.resource''
--print-executable             Print the name of the executable
--test-parser SOURCE-FILE      Dump GPD file for SOURCE-FILE to standard
                               output and exit without processing any
                               dependencies (only useful for debugging GP)
-h, --help                     Display this help and exit
--version                      Output version information and exit
... and most other options that GPC/GCC understand

Report bugs to <gpc@gnu.de>.
');
  Quit (0)
end;

procedure HandleParameters;
var
  i, j, e, Status: Integer;
  Parameter, MainSrcBase, MainSrcDir, GPExecutablePath, ExecutableName, InterfaceName, FileName: TString;
  GPCPrintOptions, ps: PStringList;
  PascalFlag: Boolean;
  OptionSelection: TOptionSelection;
  LastOption: array [Boolean, TOptionSelection] of PPOptionList;
  Parameters: array [1 .. 1] of PString;
  SelectedLanguage: (sl_None, sl_Pascal, sl_C);

  function NextParameter = Res: Boolean;
  begin
    Res := i < CmdLineOptions^.Count;
    if Res then
      begin
        Inc (i);
        Parameter := CmdLineOptions^.Elements[i]^
      end
  end;

  function OptionString (const OptionName: String; var s: String) = Res: Boolean;
  begin
    Res := IsPrefix (OptionName, Parameter);
    if Res then
      if (Length (Parameter) <= Length (OptionName)) or
         (Parameter[Length (OptionName) + 1] <> '=') then
        if OptionName[1] = '-' then
          Error ('`' + OptionName + ''' requires an argument after `=''')
        else
          Res := False  { it wasn't an option, but a file name }
      else
        s := Copy (Parameter, Length (OptionName) + 2)
  end;

  function OptionLanguage: Boolean;
  var Temp: TString;
  begin
    if OptionString ('--Pascal', Temp) or OptionString ('--pascal', Temp) then
      begin
        SelectedLanguage := sl_Pascal;
        Parameter := Temp
      end
    else if IsPrefix ('--c=', Parameter) and OptionString ('--c', Temp) then
      begin
        SelectedLanguage := sl_C;
        Parameter := Temp
      end;
    OptionLanguage := False  { always continue parsing the option }
  end;

  function OptionAppendString (const OptionName: String; var s: String; Separator: Char) = Res: Boolean;
  var Temp: TString;
  begin
    Res := OptionString (OptionName, Temp);
    if Res then s := s + Temp + Separator
  end;

  function OptionSetString (const OptionName: String; var s: String; const NewValue: String) = Res: Boolean;
  begin
    Res := IsPrefix (OptionName, Parameter);
    if Res then
      begin
        s := NewValue;
        if Length (Parameter) > Length (OptionName) then
          Warning ('ignoring invalid argument after `' + OptionName + '''')
      end
  end;

  function OptionBoolean (const OptionName: String; var v: Boolean; NewValue: Boolean) = Res: Boolean;
  begin
    Res := IsPrefix (OptionName, Parameter);
    if Res then
      begin
        v := NewValue;
        if Length (Parameter) > Length (OptionName) then
          Warning ('ignoring invalid argument after `' + OptionName + '''')
      end
  end;

  function OptionSetInteger (const OptionName: String; var v: Integer; NewValue: Integer) = Res: Boolean;
  begin
    Res := IsPrefix (OptionName, Parameter);
    if Res then
      begin
        v := NewValue;
        if Length (Parameter) > Length (OptionName) then
          Warning ('ignoring invalid argument after `' + OptionName + '''')
      end
  end;

  function OptionInteger (const OptionName: String; var v: Integer; DefaultValue: Integer) = Res: Boolean;
  var e: Integer;
  begin
    Res := IsPrefix (OptionName, Parameter);
    if Res then
      if Length (Parameter) > Length (OptionName) then
        if Parameter[Length (OptionName) + 1] <> '=' then
          Error ('argument to `' + OptionName + ''' must be after `=''')
        else
          begin
            Val (Copy (Parameter, Length (OptionName) + 2), v, e);
            if e <> 0 then
              Error ('argument to `' + OptionName + ''' must be an integer')
          end
      else
        v := DefaultValue
  end;

  function OptionObsolete (const OptionName: String) = Res: Boolean;
  begin
    Res := IsPrefix (OptionName, Parameter);
    if Res then
      Warning ('`' + OptionName + ''' is obsolete when using GP, ignoring it')
  end;

  function OptionUnsuitable (const OptionName: String) = Res: Boolean;
  begin
    Res := IsPrefix (OptionName, Parameter);
    if Res then
      Error ('`' + Parameter + ''' is meant to be used with GPC/GCC directly, not with GP')
  end;

  procedure AddParameter (const Parameter: String; MinOptionSelection: TOptionSelection; RecompileRelevant: TRecompileRelevant);
  var
    i: Integer;
    PascalOnly, COnly, PascalFlag: Boolean;
    OptionSelection: TOptionSelection;
    Par: TString;
  begin
    Par := Parameter;
    if (Length (Par) >= 2) and (Par[2] = '-') then Par[2] := 'f';
    i := 0;
    while (GPCOptions[i] <> nil) and not IsPrefix (CString2String (GPCOptions[i]), Par) do
      Inc (i);
    PascalOnly := (GPCOptions[i] <> nil) or (SelectedLanguage = sl_Pascal);
    COnly := SelectedLanguage = sl_C;
    if PascalOnly and COnly then
      Error ('option `' + Parameter + ''' is not a C option');
    for PascalFlag := PascalOnly to not COnly do
      for OptionSelection := MinOptionSelection to High (OptionSelection) do
        begin
          New (LastOption[PascalFlag, OptionSelection]^);
          Inc (Compilers[PascalFlag].OptionCount[OptionSelection]);
          LastOption[PascalFlag, OptionSelection]^^.Option := Parameter;
          LastOption[PascalFlag, OptionSelection]^^.RecompileRelevant := RecompileRelevant;
          LastOption[PascalFlag, OptionSelection] := @LastOption[PascalFlag, OptionSelection]^^.Next
        end
  end;

  function OptionWithArgument (const OptionName: String; MinOptionSelection: TOptionSelection; RecompileRelevant: TRecompileRelevant) = Res: Boolean;
  begin
    Res := IsPrefix (OptionName, Parameter);
    if Res then
      begin
        AddParameter (Parameter, MinOptionSelection, RecompileRelevant);
        if (Length (Parameter) <= Length (OptionName)) then
          if NextParameter then
            AddParameter (Parameter, MinOptionSelection, RecompileRelevant)
          else
            Error ('`' + OptionName + ''' must be followed by an argument')
      end
  end;

  function OptionWithoutArgument (const OptionName: String; MinOptionSelection: TOptionSelection; RecompileRelevant: TRecompileRelevant) = Res: Boolean;
  begin
    Res := IsPrefix (OptionName, Parameter);
    if Res then
      AddParameter (Parameter, MinOptionSelection, RecompileRelevant)
  end;

  function OptionNoWarnings = Res: Boolean;
  begin
    Res := OptionWithoutArgument ('-w', co_NoVerbose, rr_None);
    if Res then NoWarnings := True
  end;

  procedure PathSplit (Path: String; const GCCOptionName: String; MinOptionSelection: TOptionSelection; RecompileRelevant: TRecompileRelevant);
  var i: Integer;
  begin
    while Path <> '' do
      begin
        i := Pos (PathSeparator, Path);
        if i = 0 then i := Length (Path) + 1;
        if i >= 2 then
          begin
            AddParameter (GCCOptionName, MinOptionSelection, RecompileRelevant);
            AddParameter (Path[1 .. i - 1], MinOptionSelection, RecompileRelevant)
          end;
        Delete (Path, 1, i)
      end
  end;

  function OptionPathSplit (const OptionName, GCCOptionName: String; MinOptionSelection: TOptionSelection; RecompileRelevant: TRecompileRelevant) = Res: Boolean;
  var Path: TString;
  begin
    Res := OptionString (OptionName, Path);
    if Res then
      PathSplit (Path, GCCOptionName, MinOptionSelection, RecompileRelevant)
  end;

  function OptionStringList (const OptionName: String; var p: PStringList) = Res: Boolean;
  var pp: PPStringList;
  begin
    Res := IsPrefix (OptionName, Parameter);
    if Res then
      begin
        pp := @p;
        while pp^ <> nil do pp := @pp^^.Next;
        New (pp^);
        pp^^.Next := nil;
        pp^^.s := Parameter
      end
  end;

  procedure FinishCompilerOptions;
  var
    PascalFlag: Boolean;
    OptionSelection: TOptionSelection;
    p: POptionList;
  begin
    if Compilers[False].ProgName = '' then
      if Compilers[True].ProgName = '' then
        Compilers[False].ProgName := 'gcc'
      else
        Compilers[False].ProgName := Compilers[True].ProgName;
    if Compilers[True].ProgName = '' then
      Compilers[True].ProgName := 'gpc';
    LinkOptionsStr := '';
    for PascalFlag := False to True do
      with Compilers[PascalFlag] do
        begin
          for OptionSelection := Low (OptionSelection) to High (OptionSelection) do
            LastOption[PascalFlag, OptionSelection]^ := nil;
          OptionStr := '';
          p := Options[co_All];
          while p <> nil do
            begin
              if p^.RecompileRelevant = rr_Compile then
                begin
                  if OptionStr <> '' then OptionStr := OptionStr + ' ';
                  OptionStr := OptionStr + p^.Option
                end;
              if PascalFlag and (p^.RecompileRelevant = rr_Link) then
                begin
                  if LinkOptionsStr <> '' then LinkOptionsStr := LinkOptionsStr + ' ';
                  LinkOptionsStr := LinkOptionsStr + p^.Option
                end;
              p := p^.Next
            end
        end
  end;

begin
  TestParserFlag := False;
  ProgressMessages := False;
  ProgressMessageInSubprocess := False;
  ProgressBar := False;
  ProgressBarInSubprocess := False;
  Verbosity := 0;
  AutoBuild := False;
  Cleaning := 0;
  MkDirFlag := True;
  NoWarnings := False;
  DefaultPaths := True;
  CheckPlatform := True;
  CheckVersion := True;
  CheckOptions := True;
  LinkFlag := True;
  AssembleFlag := True;
  PrintOnly := False;
  PrintPlatform := False;
  PrintCycles := False;
  PrintSources := False;
  PrintObjects := False;
  PrintLibs := False;
  PrintExecutable := False;
  AutoUses := '';
  AutoUsesOption := '';
  UnitPath := GetEnv ('GPC_UNIT_PATH');
  if UnitPath <> '' then UnitPath := UnitPath + PathSeparator;
  ObjectPath := '';
  UnitDestinationPath := DefaultUnitDestinationPath;
  ObjectDestinationPath := '';
  GPExecutablePath := '';
  ExecutableName := '';
  MainSrc := '';
  MainDestName := '';
  MainDestNameGiven := False;
  GPCPrintOptions := nil;
  for PascalFlag := False to True do
    with Compilers[PascalFlag] do
      begin
        Initialized := False;
        for OptionSelection := Low (OptionSelection) to High (OptionSelection) do
          begin
            OptionCount[OptionSelection] := 0;
            LastOption[PascalFlag, OptionSelection] := @Options[OptionSelection]
          end
      end;
  { We don't use GetOpt here because then we would have to name all
    possible options to GPC and GCC, and GetOpt would allow several
    one-letter options in one argument which is not allowed for
    GPC/GCC options. So we parse the options the conventional way,
    handle those we're interested in, and pass any other options
    starting with a `-' to GPC and also to GCC unless they're known
    GPC-only options. We leave it to the compilers to complain about
    invalid options. }
  SelectedLanguage := sl_None;
  i := 0;
  if CmdLineOptions <> nil then
    while NextParameter do
      begin
        if (Parameter <> '') and (Parameter[1] = '-') then
          begin
            if (Length (Parameter) >= 2) and (Parameter[2] = 'f') then Parameter[2] := '-';
            if (Parameter = '--help') or (Parameter = '-h') then
              Usage
            else if Parameter = '--version' then
              begin
                WriteLn ('GP ' + ProgramVersion + ' (GNU Pascal Make Utility)
Copyright (C) 2000-2006 Free Software Foundation, Inc.

GP and GNU Pascal come with NO WARRANTY, to the extent permitted by
law. You may redistribute copies of GP and GNU Pascal under the terms
of the GNU General Public License. For more information about these
matters, see the file named COPYING.

Report bugs to <gpc@gnu.de>.
');
                Quit (0)
              end
            else if Parameter = '-o' then
              if NextParameter then
                begin
                  MainDestName := Parameter;
                  MainDestNameGiven := True
                end
              else
                Error ('`-o'' must be followed by the output file name')
            else if IsPrefix ('-v', Parameter) then
              if Parameter = '-v' then
                begin
                  Verbosity := MaxInt;
                  AddParameter (Parameter, co_Compile, rr_None)
                end
              else
                begin
                  Val (Copy (Parameter, 3), Verbosity, e);
                  if (e <> 0) or (Verbosity < 0) then
                    Error ('argument to `-v'' must be a nonnegative integer')
                end
            else if    { Options handled by GP and passed to the compilers }
                       OptionBoolean ('--progress-messages', ProgressMessages, True)
                    or OptionBoolean ('--no-progress-messages', ProgressMessages, False)
                    or OptionBoolean ('--progress-bar', ProgressBar, True)
                    or OptionBoolean ('--no-progress-bar', ProgressBar, False) then
              AddParameter (Parameter, co_Compile, rr_None)
            else if    { Options handled by GP and passed to the compilers and
                         not relevant for deciding whether to recompile }
                       OptionAppendString ('--unit-path', UnitPath, PathSeparator)
                    or OptionSetString ('--no-unit-path', UnitPath, '')
                    or OptionAppendString ('--object-path', ObjectPath, PathSeparator)
                    or OptionSetString ('--no-object-path', ObjectPath, '')
                    or OptionBoolean ('--disable-default-paths', DefaultPaths, False) then
              AddParameter (Parameter, co_NoVerbose, rr_None)
            else if not
                       { Options handled by GP but not (directly) passed to the compilers }
                      (OptionBoolean ('--test-parser', TestParserFlag, True)
                    or OptionBoolean ('--autobuild', AutoBuild, True)
                    or OptionBoolean ('--no-autobuild', AutoBuild, False)
                    or OptionObsolete ('--automake-gpc')
                    or OptionObsolete ('--automake-gcc')
                    or OptionObsolete ('--automake-g++')
                    or OptionBoolean ('--automake', AutoBuild, False)
                    or OptionSetInteger ('--clean-objects', Cleaning, 1)
                    or OptionSetInteger ('--clean', Cleaning, 2)
                    or OptionBoolean ('-c', LinkFlag, False)
                    or OptionBoolean ('-S', AssembleFlag, False)
                    or OptionBoolean ('--mkdir', MkDirFlag, True)
                    or OptionBoolean ('--no-mkdir', MkDirFlag, False)
                    or OptionBoolean ('--check-platform', CheckPlatform, True)
                    or OptionBoolean ('--no-check-platform', CheckPlatform, False)
                    or OptionBoolean ('--check-version', CheckVersion, True)
                    or OptionBoolean ('--no-check-version', CheckVersion, False)
                    or OptionBoolean ('--check-options', CheckOptions, True)
                    or OptionBoolean ('--no-check-options', CheckOptions, False)
                    or OptionBoolean ('--print-only', PrintOnly, True)
                    or OptionBoolean ('--print-platform', PrintPlatform, True)
                    or OptionBoolean ('--print-sources', PrintSources, True)
                    or OptionBoolean ('--print-objects', PrintObjects, True)
                    or OptionBoolean ('--print-libs', PrintLibs, True)
                    or OptionBoolean ('--print-executable', PrintExecutable, True)
                    or OptionBoolean ('--print-cycles', PrintCycles, True)
                    or OptionStringList ('--print', GPCPrintOptions)
                    or OptionStringList ('-print', GPCPrintOptions)
                    or OptionAppendString ('--uses', AutoUses, ',')
                    or OptionSetString ('--no-uses', AutoUses, '')
                    or OptionString ('--unit-destination-path', UnitDestinationPath)
                    or OptionSetString ('--no-unit-destination-path', UnitDestinationPath, DefaultUnitDestinationPath)
                    or OptionString ('--object-destination-path', ObjectDestinationPath)
                    or OptionSetString ('--no-object-destination-path', ObjectDestinationPath, '')
                    or (Parameter = '--executable-file-name')  { ignored }
                    or OptionString ('--executable-file-name', ExecutableName)
                    or OptionString ('--executable-path', GPExecutablePath)
                    or OptionSetString ('--no-executable-path', GPExecutablePath, '')
                    or OptionObsolete ('--autolink')
                    or OptionObsolete ('--no-autolink')
                    or OptionObsolete ('--no-automake')
                    or OptionObsolete ('--amtmpfile')
                    or OptionObsolete ('--gpi-destination-path')
                    or OptionUnsuitable ('--interface-only')
                    or OptionUnsuitable ('--implementation-only')
                    or OptionUnsuitable ('-x')
                    or OptionUnsuitable ('-E')
                    or OptionUnsuitable ('-M')
                    or OptionUnsuitable ('-print-')
                    or OptionUnsuitable ('--print-')
                    or OptionLanguage
                       { Options only passed to the compilers }
                    or OptionPathSplit ('--include-path', '-I', co_NoVerbose, rr_None)
                    or OptionPathSplit ('--library-path', '-L', co_NoVerbose, rr_None)
                    or OptionWithoutArgument ('-Wp', co_NoVerbose, rr_Compile)
                    or OptionWithoutArgument ('-Wa', co_Compile, rr_Compile)
                    or OptionWithoutArgument ('-Wl', co_All, rr_Link)
                    or OptionWithoutArgument ('-W', co_NoVerbose, rr_None)
                    or OptionWithoutArgument ('--time-report', co_Compile, rr_None)
                    or OptionNoWarnings
                    or OptionWithoutArgument ('-d', co_NoVerbose, rr_None)
                    or OptionWithoutArgument ('--debug', co_NoVerbose, rr_None)
                    or OptionWithoutArgument ('--save-temps', co_Compile, rr_None)
                    or OptionWithoutArgument ('-save-temps', co_Compile, rr_None)
                    or OptionWithoutArgument ('-static', co_All, rr_Link)
                    or OptionWithoutArgument ('-shared', co_All, rr_Link)
                    or OptionWithoutArgument ('-s', co_All, rr_Link)
                    or OptionWithArgument ('--include-directory', co_NoVerbose, rr_None)
                    or OptionWithArgument ('--library-directory', co_NoVerbose, rr_None)
                    or OptionWithArgument ('-B', co_OnlyVersion, rr_None)
                    or OptionWithArgument ('-b', co_OnlyVersion, rr_None)
                    or OptionWithArgument ('-V', co_OnlyVersion, rr_None)
                    or OptionWithArgument ('--define', co_NoVerbose, rr_Compile)
                    or OptionWithArgument ('-include', co_NoVerbose, rr_Compile)
                    or OptionWithArgument ('-imacros', co_NoVerbose, rr_Compile)
                    or OptionWithArgument ('-idirafter', co_NoVerbose, rr_None)
                    or OptionWithArgument ('-iprefix', co_NoVerbose, rr_None)
                    or OptionWithArgument ('-iwithprefix', co_NoVerbose, rr_None)
                    or OptionWithArgument ('-iwithprefixbefore', co_NoVerbose, rr_None)
                    or OptionWithArgument ('-isystem', co_NoVerbose, rr_None)
                    or OptionWithArgument ('--param', co_NoVerbose, rr_None)
                    or OptionWithArgument ('-A', co_NoVerbose, rr_Compile)
                    or OptionWithArgument ('-D', co_NoVerbose, rr_Compile)
                    or OptionWithArgument ('-U', co_NoVerbose, rr_Compile)
                    or OptionWithArgument ('-I', co_NoVerbose, rr_None)
                    or OptionWithArgument ('-G', co_NoVerbose, rr_Compile)
                    or OptionWithArgument ('-l', co_All, rr_Link)
                    or OptionWithArgument ('-Xlinker', co_All, rr_Link)
                    or OptionWithArgument ('-u', co_All, rr_Link)
                    or OptionWithArgument ('-L', co_All, rr_Link)) then
              AddParameter (Parameter, co_NoVerbose, rr_Compile)
          end
        else if not (   OptionString ('PC', Compilers[True].ProgName)
                     or OptionString ('CC', Compilers[False].ProgName)) then
          if MainSrc <> '' then
            Error ('only one file name allowed on the command line')
          else
            MainSrc := Parameter;
        SelectedLanguage := sl_None
      end;
  FinishCompilerOptions;
  if (MainSrc = '') and (GPCPrintOptions = nil) and not PrintPlatform then
    Usage;
  if PrintPlatform then
    begin
      InitCompiler (True);
      WriteLn (Compilers[True].Platform)
    end;
  while GPCPrintOptions <> nil do
    begin
      InitCompiler (True);
      Parameters[1] := @GPCPrintOptions^.s;
      ProgressMessage ('Doing `' + GPCPrintOptions^.s + '''');
      StartPipe (PascalFlag, co_OnlyVersion, Parameters, Null, Null, Status, Null);
      if Status <> 0 then Quit (Status);
      ps := GPCPrintOptions;
      GPCPrintOptions := GPCPrintOptions^.Next;
      Dispose (ps);
      ProgressMessage ('')
    end;
  if MainSrc = '' then Halt;
  MainSrcBase := MainSrc;
  if Ext2Language (ExtFromPath (MainSrc)) <> fk_Program then
    begin
      i := Low (LanguageSuffixes[fk_Program]);
      repeat
        MainSrc := MainSrcBase + LanguageSuffixes[fk_Program, i];
        Inc (i)
      until (i > High (LanguageSuffixes[fk_Program])) or FileExists (MainSrc)
    end;
  if not FileExists (MainSrc) then
    ErrorFile (MainSrcBase, 'file not found');
  if ExecutableName = '' then
    ExecutableName := NameFromPath (MainSrc);
  if GPExecutablePath = '' then
    GPExecutablePath := DirFromPath (MainSrc);
  if MainDestName = '' then
    MainDestName := ForceAddDirSeparator (GPExecutablePath) + ExecutableName + ExecutableSuffix;
  MainSrc := FExpand (MainSrc);
  MainSrcDir := DirFromPath (MainSrc);
  if UnitDestinationPath <> '' then
    UnitDestinationPath := ForceAddDirSeparator (FExpand (UnitDestinationPath));
  if ObjectDestinationPath = '' then
    ObjectDestinationPath := UnitDestinationPath
  else
    ObjectDestinationPath := ForceAddDirSeparator (FExpand (ObjectDestinationPath));
  UnitPath := UnitPath + UnitDestinationPath + PathSeparator + DirSelf;
  if DefaultPaths then
    UnitPath := UnitPath + PathSeparator + GetUnitPath;
  UnitPath := UnitPath + PathSeparator + MainSrcDir;
  ObjectPath := ObjectPath + ObjectDestinationPath + PathSeparator + UnitPath;
  PathSplit (ObjectPath, '-I', co_NoVerbose, rr_None);
  AddParameter ('--unit-path=' + MainSrcDir, co_NoVerbose, rr_None);
  if AutoUses <> '' then
    begin
      if AutoUses[Length (AutoUses)] = ',' then
        Delete (AutoUses, Length (AutoUses), 1);
      AutoUsesOption := '--uses=' + AutoUses;
      i := 1;
      while i <= Length (AutoUses) do
        begin
          j := i;
          while (j <= Length (AutoUses)) and
            (AutoUses[j] in ['A' .. 'Z', 'a' .. 'z', '0' .. '9', '_']) do Inc (j);
          if j <= i then
            Error ('missing interface name in `--uses'' parameter');
          InterfaceName := LoCaseStr (AutoUses[i .. j - 1]);
          FileName := '';
          if (j <= Length (AutoUses)) and (AutoUses[j] = '(') then
            begin
              Inc (j);
              i := j;
              while (j <= Length (AutoUses)) and (AutoUses[j] <> ')') do Inc (j);
              if j > Length (AutoUses) then
                Error ('missing `)'' in `--uses'' parameter');
              if j <= i then
                Error ('missing file name in `--uses'' parameter');
              FileName := AutoUses[i .. j - 1];
              Inc (j)
            end;
          i := j;
          if i < Length (AutoUses) then  { `<' so a trailing `,' will lead to an error }
            if AutoUses[i] = ',' then
              Inc (i)
            else
              Error ('missing `,'' in `--uses'' parameter');
          AutoUsedUnits := NewUsesElement (InterfaceName, FileName, AutoUsedUnits)
        end
    end;
  { We have to do this again after adding possible `-I' arguments.
    OTOH, we had to do it the first time in order to call the
    compiler and get the default unit path. }
  FinishCompilerOptions;
  { This is also not nice, but since `--uses' is not passed to those
    units themselves to avoid a self-dependency, we can't add it to
    the regular options. However, it is relevant for deciding
    whether to recompile, so we add it to OptionStr separately. }
  if AutoUsesOption <> '' then
    with Compilers[True] do
      begin
        if OptionStr <> '' then OptionStr := OptionStr + ' ';
        OptionStr := OptionStr + AutoUsesOption
      end;
  NeededOptions := GetNeededOptions;
  if (Verbosity >= 2) and (NeededOptions <> nil) then
    for e := 1 to NeededOptions^.Count do
      Message (2, 'Needed option for GPC: ' + NeededOptions^[e]^)
end;

var
  Deps: PDepList;

begin
  ProgramName := ParamStr (0);
  GPConfigInit (Compilers[True].ProgName, Compilers[False].ProgName, ExecutableSuffix, CmdLineOptions);
  HandleParameters;
  if TestParserFlag then
    begin
      TestParser (MainSrc);
      Quit (0)
    end;
  ProgressBarSet (0);
  Deps := CollectDeps (MainSrc);
  if Cleaning <> 0 then
    begin
      DoClean (Deps, Cleaning);
      ProgressMessage ('Cleaning done.');
      ProgressBarSet (1);
      Halt
    end;
  if not PrintOnly then
    begin
      ProcessDependencies (Deps);
      if AssembleFlag then
        if LinkFlag then
          Link (Deps)
        else if MainDestNameGiven then
          Warning ('ignoring `-o'' when not linking');
      ProgressMessage ('Compilation completed.')
    end;
  DoPrint (Deps)
end.
