/*Definitions for switches for Pascal.

  Copyright (C) 1997-2005 Free Software Foundation, Inc.

  Authors: Peter Gerwinski <peter@gerwinski.de>
           Frank Heckenbach <frank@pascal.gnu.de>

  This file is part of GNU Pascal.

  GNU Pascal is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published
  by the Free Software Foundation; either version 2, or (at your
  option) any later version.

  GNU Pascal is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GNU Pascal; see the file COPYING. If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA. */

/* This is the contribution to the `lang_options' array in gcc.c for gpc. */

#ifndef OPTIONS_ONLY
#include "gcc-version.h"
#ifdef EGCS
DEFINE_LANG_NAME ("Pascal")
#define GPC_OPT0(DIRECTIVE, NAME, DESCRIPTION) { NAME, DESCRIPTION },
#define GPC_OPT1(DIRECTIVE, NAME, SETTING, DESCRIPTION) { NAME, DESCRIPTION },
#else
#define GPC_OPT0(DIRECTIVE, NAME, DESCRIPTION) NAME,
#define GPC_OPT1(DIRECTIVE, NAME, SETTING, DESCRIPTION) NAME,
#endif
#endif

/* Note: If the name of option A is a prefix of the name of option B, A must come *after* B. */
  GPC_OPT0 (0, "-famtmpfile", "(Internal switch used for automake)")
  GPC_OPT1 (1, "-fassertions", assertions = 1, "Enable assertion checking (default)")
  GPC_OPT1 (1, "-fno-assertions", assertions = 0, "Inverse of `--assertions'")
  GPC_OPT1 (0, "-fautobuild", automake_level = 3, "Automatically compile all units/modules/`{$L ...}' files and link the object files provided")
  GPC_OPT1 (0, "-fno-autobuild", automake_level = 0, "Inverse of `--autobuild'")
  GPC_OPT1 (0, "-fautolink", automake_level = 1, "Automatically link object files provided by units/modules or `{$L ...}' (default)")
  GPC_OPT1 (0, "-fno-autolink", automake_level = 0, "Inverse of `--autolink'")
  GPC_OPT0 (0, "-fautomake-gcc", "Set the C compiler invoked by automake")
  GPC_OPT0 (0, "-fautomake-g++", "Set the C++ compiler invoked by automake")
  GPC_OPT0 (0, "-fautomake-gpc", "Set the Pascal compiler invoked by automake")
  GPC_OPT1 (0, "-fautomake", automake_level = 2, "Automatically compile changed units/modules/`{$L ...}' files and link the object files provided")
  GPC_OPT1 (0, "-fno-automake", automake_level = 0, "Inverse of `--automake'")
  GPC_OPT1 (0, "-fbig-endian", option_big_endian = 1, "Tell GPC that the system is big-endian (for those targets where it can vary)")
  GPC_OPT1 (1, "-fborland-pascal", pascal_dialect = BORLAND_PASCAL, "Try to emulate Borland Pascal, version 7.0")
  GPC_OPT1 (1, "-fcase-value-checking", case_value_checking = 1, "Cause a runtime error if a `case' matches no branch (default in ISO Pascal modes)")
  GPC_OPT1 (1, "-fno-case-value-checking", case_value_checking = 0, "Inverse of `--case-value-checking'")
  GPC_OPT0 (0, "-fcidefine", /* handled specially by the preprocessor when in the source */ "Define a case-insensitive macro")
  GPC_OPT1 (1, "-fclassic-pascal-level-0", pascal_dialect = CLASSIC_PASCAL_LEVEL_0, "Reject conformant arrays and anything besides ISO 7185 Pascal")
  GPC_OPT1 (1, "-fclassic-pascal", pascal_dialect = CLASSIC_PASCAL_LEVEL_1, "Reject anything besides ISO 7185 Pascal")
  GPC_OPT0 (0, "-fcsdefine", /* handled specially by the preprocessor when in the source */ "Define a case-sensitive macro")
  GPC_OPT1 (1, "-fcstrings-as-strings", cstrings_as_strings = 1, "Treat CStrings as strings")
  GPC_OPT1 (1, "-fno-cstrings-as-strings", cstrings_as_strings = 0, "Inverse of `--cstrings-as-strings'")
  GPC_OPT1 (1, "-fdebug-automake", debug_automake = 1, "(For GPC developers.) Give additional information about the actions of automake")
  GPC_OPT1 (1, "-fdebug-gpi", debug_gpi = 1, "(For GPC developers.) Show what is written to and read from GPI files (huge output!)")
  GPC_OPT1 (1, "-fdebug-source", debug_source = 1, "Output the source while it is being processed to standard error")
  GPC_OPT1 (1, "-fno-debug-source", debug_source = 0, "Inverse of `--debug-source'")
  GPC_OPT0 (1, "-fdebug-tree", "(For GPC developers.) Show the internal representation of a given tree node (name or address)")
  GPC_OPT1 (1, "-fdelphi-comments", delphi_comments = 1, "Allow Delphi style `//' comments (default)")
  GPC_OPT1 (1, "-fno-delphi-comments", delphi_comments = 0, "Inverse of `--delphi-comments'")
  GPC_OPT1 (1, "-fdelphi", pascal_dialect = BORLAND_DELPHI, "Try to emulate Borland Pascal, version 7.0, with some Delphi extensions")
  GPC_OPT0 (1, "-fdisable-debug-info", "Inhibit `-g' options (temporary work-around, this option may disappear in the future)")
  GPC_OPT0 (0, "-fdisable-default-paths", "Do not add a default path to the unit and object path")
  GPC_OPT0 (1, "-fdisable-keyword", "Disable a keyword, independently of dialect defaults")
  GPC_OPT0 (1, "-fdisable-predefined-identifier", "Disable a predefined identifier, independently of dialect defaults")
  GPC_OPT1 (1, "-fdouble-quoted-strings", double_quoted_strings = 1, "Allow strings enclosed in \"\" (default with `--gnu-pascal' and `--mac-pascal')")
  GPC_OPT1 (1, "-fno-double-quoted-strings", double_quoted_strings = 0, "Inverse of `--double-quoted-strings'")
  GPC_OPT0 (1, "-fenable-keyword", "Enable a keyword, independently of dialect defaults")
  GPC_OPT0 (1, "-fenable-predefined-identifier", "Enable a predefined identifier, independently of dialect defaults")
  GPC_OPT1 (1, "-fexact-compare-strings", exact_compare_strings = 1, "Do not blank-pad strings for comparisons")
  GPC_OPT1 (1, "-fno-exact-compare-strings", exact_compare_strings = 0, "Inverse of `--exact-compare-strings'")
  GPC_OPT0 (0, "-fexecutable-file-name", "Name for the output file, if specified; otherwise derive from main source file name")
  GPC_OPT0 (0, "-fexecutable-path", "Path where to create the executable file")
  GPC_OPT0 (0, "-fno-executable-path", "Create the executable file in the directory where the main source is (default)")
  GPC_OPT1 (1, "-fextended-pascal", pascal_dialect = EXTENDED_PASCAL, "Reject anything besides ISO 10206 Extended Pascal")
  GPC_OPT0 (1, "-fextended-syntax", "same as `--ignore-function-results --pointer-arithmetic --cstrings-as-strings -Wno-absolute' (same as `{$X+}')")
  GPC_OPT0 (1, "-fno-extended-syntax", "Inverse of `--extended-syntax' (same as `{$X-}')")
  GPC_OPT0 (1, "-ffield-widths", "Optional colon-separated list of default field widths for Integer, Real, Boolean, LongInt, LongReal")
  GPC_OPT0 (1, "-fno-field-widths", "Inverse of `--field-widths'")
  GPC_OPT1 (1, "-fgnu-pascal", pascal_dialect = 0, "Undo the effect of previous dialect options, allow all default features again")
  GPC_OPT0 (1, "-fgpc-main", "External name for the program's entry point (default: `main')")
  GPC_OPT0 (0, "-fgpi-destination-path", "(Internal switch used for automake)")
  GPC_OPT1 (1, "-fignore-function-results", ignore_function_results = 1, "Do not complain when a function is called like a procedure")
  GPC_OPT1 (1, "-fno-ignore-function-results", ignore_function_results = 0, "Inverse of `--ignore-function-results'")
  GPC_OPT1 (1, "-fignore-garbage-after-dot", ignore_garbage_after_dot = 1, "Ignore anything after the terminating `.' (default in `--borland-pascal')")
  GPC_OPT1 (1, "-fno-ignore-garbage-after-dot", ignore_garbage_after_dot = 0, "Inverse of `--ignore-garbage-after-dot'")
  GPC_OPT1 (1, "-fignore-packed", ignore_packed = 1, "Ignore `packed' in the source code (default in `--borland-pascal')")
  GPC_OPT1 (1, "-fno-ignore-packed", ignore_packed = 0, "Inverse of `--ignore-packed'")
  GPC_OPT0 (0, "-fimplementation-only", "Do not produce a GPI file; only compile the implementation part")
  GPC_OPT1 (1, "-fimplicit-result", implicit_result = 1, "Enable implicit `Result' for functions (default only in `--delphi')")
  GPC_OPT1 (1, "-fno-implicit-result", implicit_result = 0, "Inverse of `--implicit-result'")
  GPC_OPT0 (1, "-finit-modules", "Initialize the named modules in addition to those imported regularly; kind of a kludge")
  GPC_OPT1 (0, "-finterface-only", interface_only = 1, "Compile only the interface part of a unit/module and exit (creates `.gpi' file, no `.o' file")
  GPC_OPT1 (1, "-fio-checking", io_checking = 1, "Check I/O operations automatically (same as `{$I+}') (default)")
  GPC_OPT1 (1, "-fno-io-checking", io_checking = 0, "Inverse of `--io-checking' (same as `{$I-}')")
  GPC_OPT1 (1, "-fiso-gotos", iso_gotos = 1, "Do not allow gotos into structured statements (default in ISO Pascal modes)")
  GPC_OPT1 (1, "-fno-iso-gotos", iso_gotos = 0, "Inverse of `--iso-gotos'")
  GPC_OPT1 (0, "-flittle-endian", option_big_endian = 0, "Tell GPC that the system is little-endian (for those targets where it can vary)")
  GPC_OPT1 (1, "-flongjmp-all-nonlocal-labels", longjmp_all_nonlocal_labels = 1, "Use `longjmp' for all nonlocal labels, not only those in the main program")
  GPC_OPT1 (1, "-fno-longjmp-all-nonlocal-labels", longjmp_all_nonlocal_labels = 0, "Inverse of `--longjmp-all-nonlocal-labels'")
  GPC_OPT1 (1, "-fmac-pascal", pascal_dialect = MAC_PASCAL, "Support (some features of) traditional Macintosh Pascal compilers")
  GPC_OPT0 (1, "-fmacros", "Expand macros (default, except with `--ucsd-pascal', `--borland-pascal' or `--delphi')")
  GPC_OPT0 (1, "-fno-macros", "Inverse of `--macros'")
  GPC_OPT0 (1, "-fmaximum-field-alignment", "Set the maximum field alignment in bits if `pack-struct' is in effect")
  GPC_OPT1 (1, "-fmethods-always-virtual", methods_always_virtual = 1, "Make all methods virtual (default in `--mac-pascal')")
  GPC_OPT1 (1, "-fno-methods-always-virtual", methods_always_virtual = 0, "Inverse of `--methods-always-virtual'")
  GPC_OPT1 (1, "-fmixed-comments", mixed_comments = 1, "Allow comments like `{ ... *)' as required in ISO Pascal (default in ISO 7185/10206 Pascal mode)")
  GPC_OPT1 (1, "-fno-mixed-comments", mixed_comments = 0, "Inverse of `--mixed-comments': Ignore `{' and `}' within `(* ... *)' comments and vice versa (default)")
  GPC_OPT1 (1, "-fnested-comments", nested_comments = 1, "Allow nested comments like `{ { } }' and `(* (* *) *)'")
  GPC_OPT1 (1, "-fno-nested-comments", nested_comments = 0, "Inverse of `--nested-comments'")
  GPC_OPT1 (1, "-fnonlocal-exit", nonlocal_exit = 1, "Allow non-local `Exit' statements (default in `--ucsd-pascal' and `--mac-pascal')")
  GPC_OPT1 (1, "-fno-nonlocal-exit", nonlocal_exit = 0, "Inverse of `--nonlocal-exit'")
  GPC_OPT1 (1, "-fno-object-checking", object_checking = 0, "Inverse of `--object-checking'")
  GPC_OPT1 (1, "-fobject-checking", object_checking = 1, "Check for valid objects on virtual method calls (default)")
  GPC_OPT0 (0, "-fobject-destination-path", "Path where to create additional object files (e.g. of C files, not Pascal units)")
  GPC_OPT0 (0, "-fno-object-destination-path", "Create additional object files (e.g. of C files, not Pascal units) in the current directory (default)")
  GPC_OPT1 (1, "-fobject-pascal", pascal_dialect = OBJECT_PASCAL, "Reject anything besides (the implemented parts of) ANSI draft Object Pascal")
  GPC_OPT0 (0, "-fobject-path", "Directories where to look for additional object (and source) files")
  GPC_OPT0 (0, "-fno-object-path", "Forget about directories where to look for additional object (and source) files")
  GPC_OPT1 (1, "-fpascal-sc", pascal_dialect = PASCAL_SC, "Be strict about the implemented Pascal-SC extensions")
  GPC_OPT0 (1, "-fpedantic", "Reject everything not allowed in some dialect, e.g. redefinition of its keywords")
  GPC_OPT0 (1, "-fno-pedantic", "Inverse of `--pedantic'")
  GPC_OPT1 (1, "-fpointer-arithmetic", pointer_arithmetic = 1, "Enable pointer arithmetic")
  GPC_OPT1 (1, "-fno-pointer-arithmetic", pointer_arithmetic = 0, "Inverse of `--pointer-arithmetic'")
  GPC_OPT1 (1, "-fpointer-checking-user-defined", pointer_checking_user_defined = 1, "Use user-defined procedure for validating pointers")
  GPC_OPT1 (1, "-fno-pointer-checking-user-defined", pointer_checking_user_defined = 0, "Inverse of `--pointer-checking-user-defined'")
  GPC_OPT1 (1, "-fpointer-checking", pointer_checking = 1, "Validate pointers before dereferencing")
  GPC_OPT1 (1, "-fno-pointer-checking", pointer_checking = 0, "Inverse of `--pointer-checking'")
  GPC_OPT1 (0, "-fprint-needed-options", print_needed_options = 1, "Print the needed options")
  GPC_OPT0 (0, "-fprogress-bar", "Output number of processed lines while compiling")
  GPC_OPT0 (0, "-fno-progress-bar", "Inverse of `--progress-bar'")
  GPC_OPT0 (0, "-fprogress-messages", "Output source file names and line numbers while compiling")
  GPC_OPT0 (0, "-fno-progress-messages", "Inverse of `--progress-messages'")
  GPC_OPT1 (1, "-fpropagate-units", propagate_units = 1, "Automatically re-export all imported declarations")
  GPC_OPT1 (1, "-fno-propagate-units", propagate_units = 0, "Inverse of `--propagate-units'")
  GPC_OPT0 (1, "-frange-and-object-checking", "Same as `--range-checking --object-checking', same as `{$R+}'")
  GPC_OPT0 (1, "-fno-range-and-object-checking", "Inverse of `--range-and-object-checking' (same as `{$R-}'")
  GPC_OPT1 (1, "-frange-checking", range_checking = 1, "Do automatic range checks') (default)")
  GPC_OPT1 (1, "-fno-range-checking", range_checking = 0, "Inverse of `--range-checking'")
  GPC_OPT1 (1, "-fread-base-specifier", read_base_specifier = 1, "In read statements, allow input base specifier `n#' (default, except in ISO 7185 Pascal)")
  GPC_OPT1 (1, "-fno-read-base-specifier", read_base_specifier = 0, "Inverse of `--read-base-specifier'")
  GPC_OPT1 (1, "-fread-hex", read_hex = 1, "In read statements, allow hexadecimal input with `$' (default, except in ISO Pascal modes)")
  GPC_OPT1 (1, "-fno-read-hex", read_hex = 0, "Inverse of `--no-read-hex'")
  GPC_OPT1 (1, "-fread-white-space", read_white_space = 1, "In read statements, require whitespace after numbers")
  GPC_OPT1 (1, "-fno-read-white-space", read_white_space = 0, "Inverse of `--read-white-space'")
  GPC_OPT0 (1, "-fsetlimit", "Define the range for `set of Integer' etc.")
  GPC_OPT1 (1, "-fshort-circuit", short_circuit = 1, "Guarantee short-circuit Boolean evaluation (default; same as `{$B-}')")
  GPC_OPT1 (1, "-fno-short-circuit", short_circuit = 0, "Inverse of `--short-circuit' (same as `{$B+}')")
  GPC_OPT0 (1, "-fstack-checking", "Enable stack checking (same as `{$S+}')")
  GPC_OPT0 (1, "-fno-stack-checking", "Inverse of `--stack-checking' (same as `{$S-}')")
  GPC_OPT1 (1, "-fstandard-pascal-level-0", pascal_dialect = CLASSIC_PASCAL_LEVEL_0, "Synonym for `--classic-pascal-level-0'")
  GPC_OPT1 (1, "-fstandard-pascal", pascal_dialect = CLASSIC_PASCAL_LEVEL_1, "Synonym for `--classic-pascal'")
  GPC_OPT1 (1, "-fsun-pascal", pascal_dialect = SUN_PASCAL, "Support (a few features of) Sun Pascal")
  GPC_OPT1 (1, "-ftransparent-file-names", transparent_file_names = 1, "Derive external file names from variable names")
  GPC_OPT1 (1, "-fno-transparent-file-names", transparent_file_names = 0, "Inverse of `--transparent-file-names'")
  GPC_OPT1 (1, "-ftruncate-strings", truncate_strings = 1, "Truncate strings being assigned to other strings of too short capacity")
  GPC_OPT1 (1, "-fno-truncate-strings", truncate_strings = 0, "Inverse of `--truncate-strings'")
  GPC_OPT1 (1, "-ftyped-address", typed_address = 1, "Make the result of the address operator a typed pointer (same as `{$T+}', default)")
  GPC_OPT1 (1, "-fno-typed-address", typed_address = 0, "Inverse of `--typed-address' (same as `{$T-}')")
  GPC_OPT1 (1, "-fucsd-pascal", pascal_dialect = UCSD_PASCAL, "Try to emulate UCSD Pascal")
  GPC_OPT0 (0, "-funit-destination-path", "Path where to create object and GPI files of Pascal units")
  GPC_OPT0 (0, "-fno-unit-destination-path", "Create object and GPI files of Pascal units in the current directory (default)")
  GPC_OPT0 (0, "-funit-path", "Directories where to look for unit/module sources")
  GPC_OPT0 (0, "-fno-unit-path", "Forget about directories where to look for unit/module sources")
  GPC_OPT1 (1, "-funtyped-pointers", untyped_pointers = 1, "Allow working (dereferencing, arithmetic) with untyped pointers")
  GPC_OPT1 (1, "-fno-untyped-pointers", untyped_pointers = 0, "Inverse of `--untyped-pointers'")
  GPC_OPT0 (0, "-fuses", "Add an implicit `uses' clause")
  GPC_OPT1 (1, "-fvax-pascal", pascal_dialect = VAX_PASCAL, "Support (a few features of) VAX Pascal")
  GPC_OPT1 (1, "-fwrite-capital-exponent", capital_exponent = 1, "Write real exponents with a capital `E'")
  GPC_OPT1 (1, "-fno-write-capital-exponent", capital_exponent = 0, "Inverse of `--write-capital-exponent'")
  GPC_OPT1 (1, "-fwrite-clip-strings", write_clip_strings = 1, "In write statements, truncate strings exceeding their field width (`Write (SomeLongString : 3)')")
  GPC_OPT1 (1, "-fno-write-clip-strings", write_clip_strings = 0, "Inverse of `--write-clip-strings'")
  GPC_OPT1 (1, "-fwrite-real-blank", real_blank = 1, "Output a blank in front of positive reals in exponential form (default)")
  GPC_OPT1 (1, "-fno-write-real-blank", real_blank = 0, "Inverse of `--write-real-blank'")
  GPC_OPT1 (1, "-Wabsolute", warn_absolute = 1, "Warn about variables at absolute adresses and `absolute' variable with non-constant addresses (default)")
  GPC_OPT1 (1, "-Wno-absolute", warn_absolute = 0, "Inverse of `-Wabsolute'")
  GPC_OPT1 (1, "-Wdynamic-arrays", warn_dynamic_arrays = 1, "Warn about arrays whose size is determined at run time (including array slices)")
  GPC_OPT1 (1, "-Wno-dynamic-arrays", warn_dynamic_arrays = 0, "Inverse of `-Wdynamic-arrays'")
#ifndef EGCS97
  /* gcc-3.x supports the following two options already. Listing them here
     would prevent automake from passing them to C compiler invocations. */
  GPC_OPT0 (1, "-Wfloat-equal", "Warn about `=' and `<>' comparisons of real numbers")
  GPC_OPT0 (1, "-Wno-float-equal", "Inverse of `-Wfloat-equal'")
#endif
  GPC_OPT1 (1, "-Widentifier-case-local", warn_id_case = 1, "Warn about an identifier written with varying case within one program/module/unit")
  GPC_OPT1 (1, "-Wno-identifier-case-local", warn_id_case = 0, "Inverse of `-Widentifier-case-local'")
  GPC_OPT1 (1, "-Widentifier-case", warn_id_case = 2, "Warn about an identifier written with varying case")
  GPC_OPT1 (1, "-Wno-identifier-case", warn_id_case = 0, "Inverse of `-Widentifier-case'")
  GPC_OPT1 (1, "-Wimplicit-abstract", warn_implicit_abstract = 1, "Warn when an object type not declared `abstract' contains an abstract method (default)")
  GPC_OPT1 (1, "-Wno-implicit-abstract", warn_implicit_abstract = 0, "Inverse of `-Wimplicit-abstract'")
  GPC_OPT1 (1, "-Wimplicit-io", warn_implicit_io = 1, "Warn when `Input' or `Output' are used implicitly")
  GPC_OPT1 (1, "-Wno-implicit-io", warn_implicit_io = 0, "Inverse of `-Wimplicit-io'")
  GPC_OPT1 (1, "-Winherited-abstract", warn_inherited_abstract = 1, "Warn when an abstract object type inherits from a non-abstract one (default)")
  GPC_OPT1 (1, "-Wno-inherited-abstract", warn_inherited_abstract = 0, "Inverse of `-Winherited-abstract'")
  GPC_OPT1 (1, "-Winterface-file-name", warn_interface_file_name = 1, "Warn when a unit/module interface differs from the file name")
  GPC_OPT1 (1, "-Wno-interface-file-name", warn_interface_file_name = 0, "Inverse of `-Winterface-file-name'")
  GPC_OPT1 (1, "-Wlocal-external", warn_local_external = 1, "Warn about local `external' declarations")
  GPC_OPT1 (1, "-Wno-local-external", warn_local_external = 0, "Inverse of `-Wlocal-external'")
  GPC_OPT1 (1, "-Wmixed-comments", warn_mixed_comments = 1, "Warn about mixed comments like `{ ... *)'")
  GPC_OPT1 (1, "-Wno-mixed-comments", warn_mixed_comments = 0, "Inverse of `-Wmixed-comments'")
  GPC_OPT1 (1, "-Wnear-far", warn_near_far = 1, "Warn about use of useless `near' or `far' directives (default)")
  GPC_OPT1 (1, "-Wno-near-far", warn_near_far = 0, "Inverse of `-Wnear-far'")
  GPC_OPT1 (1, "-Wnested-comments", warn_nested_comments = 1, "Warn about nested comments like `{ { } }'")
  GPC_OPT1 (1, "-Wno-nested-comments", warn_nested_comments = 0, "Inverse of `-Wnested-comments: Warn about nested comments'")
  GPC_OPT1 (1, "-Wobject-assignment", warn_object_assignment = 1, "Warn when when assigning objects or declaring them as value parameters or function results (default, except in `--borland-pascal')")
  GPC_OPT1 (1, "-Wno-object-assignment", warn_object_assignment = 0, "Inverse of `-Wobject-assignment'")
  GPC_OPT1 (1, "-Wsemicolon", warn_semicolon = 1, "Warn about a semicolon after `then', `else' or `do' (default)")
  GPC_OPT1 (1, "-Wno-semicolon", warn_semicolon = 0, "Inverse of `-Wsemicolon'")
  GPC_OPT1 (1, "-Wtyped-const", warn_typed_const = 1, "Warn about misuse of typed constants as initialized variables (default)")
  GPC_OPT1 (1, "-Wno-typed-const", warn_typed_const = 0, "Inverse of `-Wtyped-const'")
  GPC_OPT1 (1, "-Wunderscore", warn_underscore = 1, "Warn about double/leading/trailing underscores in identifiers")
  GPC_OPT1 (1, "-Wno-underscore", warn_underscore = 0, "Inverse of `-Wunderscore'")
  GPC_OPT0 (1, "-Wwarnings", "Enable warnings (same as `{$W+}')")
  GPC_OPT0 (1, "-Wno-warnings", "Inverse of `-Wwarnings' (same as `{$W-}')")
